@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -22,7 +22,19 @@
 
 #include "system.h"
 
-#include <event.h>
+#include <usual/time.h>
+#include <usual/list.h>
+#include <usual/statlist.h>
+#include <usual/string.h>
+#include <usual/logging.h>
+#include <usual/aatree.h>
+#include <usual/lookup3.h>
+#include <usual/slab.h>
+#include <usual/socket.h>
+#include <usual/safeio.h>
+#include <usual/mbuf.h>
+#include <usual/event.h>
+#include <usual/strpool.h>
 
 #ifdef DBGVER
 #define FULLVER   PACKAGE_NAME " version " PACKAGE_VERSION " (" DBGVER ")"
@@ -62,20 +74,18 @@ typedef struct PgUser PgUser;
 typedef struct PgDatabase PgDatabase;
 typedef struct PgPool PgPool;
 typedef struct PgStats PgStats;
-typedef struct PgAddr PgAddr;
+typedef union PgAddr PgAddr;
 typedef enum SocketState SocketState;
 typedef struct PktHdr PktHdr;
 
-#include "aatree.h"
-#include "hash.h"
+extern int cf_sbuf_len;
+
 #include "util.h"
-#include "list.h"
-#include "mbuf.h"
 #include "iobuf.h"
 #include "sbuf.h"
 #include "pktbuf.h"
 #include "varcache.h"
-#include "slab.h"
+#include "dnslookup.h"
 
 #include "admin.h"
 #include "loader.h"
@@ -102,6 +112,7 @@ typedef struct PktHdr PktHdr;
 #define AUTH_CREDS	6
 
 /* type codes for weird pkts */
+#define PKT_STARTUP_V2  0x20000
 #define PKT_STARTUP     0x30000
 #define PKT_CANCEL      80877102
 #define PKT_SSLREQ      80877103
@@ -117,15 +128,36 @@ typedef struct PktHdr PktHdr;
 
 #define BACKENDKEY_LEN	8
 
+/* buffer size for startup noise */
+#define STARTUP_BUF	1024
+
+
 /*
  * Remote/local address
  */
-struct PgAddr {
-	struct in_addr ip_addr;
-	unsigned short port;
-	bool is_unix;
+
+/* buffer for pgaddr string conversions (with port) */
+#define PGADDR_BUF  (INET6_ADDRSTRLEN + 10)
+
+/*
+ * AF_INET,AF_INET6 are stored as-is,
+ * AF_UNIX uses sockaddr_in port.
+ */
+union PgAddr {
+	struct sockaddr sa;
+	struct sockaddr_in sin;
+	struct sockaddr_in6 sin6;
 };
 
+static inline bool pga_is_unix(const PgAddr *a) { return a->sa.sa_family == AF_UNIX; }
+
+int pga_port(const PgAddr *a);
+void pga_set(PgAddr *a, int fam, int port);
+void pga_copy(PgAddr *a, const struct sockaddr *sa);
+bool pga_pton(PgAddr *a, const char *s, int port);
+const char *pga_ntop(const PgAddr *a, char *dst, int dstlen);
+const char *pga_str(const PgAddr *a, char *dst, int dstlen);
+
 /*
  * Stats, kept per-pool.
  */
@@ -146,29 +178,28 @@ struct PgStats {
  *   ->newer_stats = ->stats
  */
 struct PgPool {
-	List head;			/* entry in global pool_list */
-	List map_head;			/* entry in user->pool_list */
+	struct List head;			/* entry in global pool_list */
+	struct List map_head;			/* entry in user->pool_list */
 
 	PgDatabase *db;			/* corresponging database */
 	PgUser *user;			/* user logged in as */
 
-	StatList active_client_list;	/* waiting events logged in clients */
-	StatList waiting_client_list;	/* client waits for a server to be available */
-	StatList cancel_req_list;	/* closed client connections with server key */
+	struct StatList active_client_list;	/* waiting events logged in clients */
+	struct StatList waiting_client_list;	/* client waits for a server to be available */
+	struct StatList cancel_req_list;	/* closed client connections with server key */
 
-	StatList active_server_list;	/* servers linked with clients */
-	StatList idle_server_list;	/* servers ready to be linked with clients */
-	StatList used_server_list;	/* server just unlinked from clients */
-	StatList tested_server_list;	/* server in testing process */
-	StatList new_server_list;	/* servers in login phase */
+	struct StatList active_server_list;	/* servers linked with clients */
+	struct StatList idle_server_list;	/* servers ready to be linked with clients */
+	struct StatList used_server_list;	/* server just unlinked from clients */
+	struct StatList tested_server_list;	/* server in testing process */
+	struct StatList new_server_list;	/* servers in login phase */
 
 	PgStats stats;
 	PgStats newer_stats;
 	PgStats older_stats;
 
 	/* database info to be sent to client */
-	uint8_t welcome_msg[256];	/* ServerParams without VarCache ones */
-	unsigned welcome_msg_len;
+	struct PktBuf *welcome_msg; /* ServerParams without VarCache ones */
 
 	VarCache orig_vars;		/* default params from server */
 
@@ -204,9 +235,9 @@ struct PgPool {
  * whis user has logged in.
  */
 struct PgUser {
-	List head;		/* used to attach user to list */
-	List pool_list;		/* list of pools where pool->user == this user */
-	Node tree_node;		/* used to attach user to tree */
+	struct List head;		/* used to attach user to list */
+	struct List pool_list;		/* list of pools where pool->user == this user */
+	struct AANode tree_node;	/* used to attach user to tree */
 	char name[MAX_USERNAME];
 	char passwd[MAX_PASSWORD];
 };
@@ -215,27 +246,31 @@ struct PgUser {
  * A database entry from config.
  */
 struct PgDatabase {
-	List head;
+	struct List head;
 	char name[MAX_DBNAME];	/* db name for clients */
 
 	bool db_paused;		/* PAUSE <db>; was issued */
 	bool db_dead;		/* used on RELOAD/SIGHUP to later detect removed dbs */
+	bool db_auto;		/* is the database auto-created by autodb_connstr */
 	bool admin;		/* internal console db */
 
-	uint8_t startup_params[256]; /* partial StartupMessage (without user) be sent to server */
-	unsigned startup_params_len;
+	struct PktBuf *startup_params; /* partial StartupMessage (without user) be sent to server */
 
 	PgUser *forced_user;	/* if not NULL, the user/psw is forced */
 
-	PgAddr addr;		/* address prepared for connect() */
-	char unix_socket_dir[UNIX_PATH_MAX]; /* custom unix socket dir */
+	const char *host;	/* host or unix socket name */
+	int port;
 
 	int pool_size;		/* max server connections in one pool */
+	int res_pool_size;	/* additional server connections in case of trouble */
 
 	const char *dbname;	/* server-side name, pointer to inside startup_msg */
 
 	/* startup commands to send to server after connect. malloc-ed */
 	const char *connect_query;
+
+	usec_t inactive_time;	/* when auto-database became inactive (to kill it after timeout) */
+	unsigned active_stamp;	/* set if autodb has connections */
 };
 
 
@@ -245,7 +280,7 @@ struct PgDatabase {
  * ->state corresponds to various lists the struct can be at.
  */
 struct PgSocket {
-	List head;		/* list header */
+	struct List head;		/* list header */
 	PgSocket *link;		/* the dest of packets */
 	PgPool *pool;		/* parent pool, if NULL not yet assigned */
 
@@ -274,6 +309,8 @@ struct PgSocket {
 	PgAddr remote_addr;	/* ip:port for remote endpoint */
 	PgAddr local_addr;	/* ip:port for local endpoint */
 
+	struct DNSToken *dns_token;	/* ongoing request */
+
 	VarCache vars;		/* state of interesting server parameters */
 
 	SBuf sbuf;		/* stream buffer, must be last */
@@ -291,21 +328,24 @@ struct PgSocket {
 #define tmp_login_salt  cancel_key
 
 /* main.c */
-extern int cf_verbose;
 extern int cf_daemon;
-extern int cf_quiet;
 
+extern char *cf_config_file;
 extern char *cf_jobname;
-extern int cf_syslog;
-extern char *cf_syslog_facility;
 
 extern char *cf_unix_socket_dir;
 extern char *cf_listen_addr;
 extern int cf_listen_port;
+extern int cf_listen_backlog;
 
 extern int cf_pool_mode;
 extern int cf_max_client_conn;
 extern int cf_default_pool_size;
+extern int cf_res_pool_size;
+extern usec_t cf_res_pool_timeout;
+
+extern char * cf_autodb_connstr;
+extern usec_t cf_autodb_idle_timeout;
 
 extern usec_t cf_suspend_timeout;
 extern usec_t cf_server_lifetime;
@@ -316,14 +356,16 @@ extern usec_t cf_server_check_delay;
 extern usec_t cf_server_connect_timeout;
 extern usec_t cf_server_login_retry;
 extern usec_t cf_query_timeout;
+extern usec_t cf_query_wait_timeout;
 extern usec_t cf_client_idle_timeout;
 extern usec_t cf_client_login_timeout;
 extern int cf_server_round_robin;
+extern int cf_disable_pqexec;
+extern usec_t cf_dns_max_ttl;
 
 extern int cf_auth_type;
 extern char *cf_auth_file;
 
-extern char *cf_logfile;
 extern char *cf_pidfile;
 
 extern char *cf_ignore_startup_params;
@@ -336,7 +378,7 @@ extern int cf_pause_mode;
 extern int cf_shutdown;
 extern int cf_reboot;
 
-extern int cf_sbuf_len;
+extern int cf_sbuf_loopcnt;
 extern int cf_tcp_keepalive;
 extern int cf_tcp_keepcnt;
 extern int cf_tcp_keepidle;
@@ -348,27 +390,31 @@ extern int cf_log_connections;
 extern int cf_log_disconnections;
 extern int cf_log_pooler_errors;
 
-extern ConfElem bouncer_params[];
-
 extern usec_t g_suspend_start;
 
+extern struct DNSContext *adns;
+
 static inline PgSocket * _MUSTCHECK
-pop_socket(StatList *slist)
+pop_socket(struct StatList *slist)
 {
-	List *item = statlist_pop(slist);
+	struct List *item = statlist_pop(slist);
 	if (item == NULL)
 		return NULL;
 	return container_of(item, PgSocket, head);
 }
 
 static inline PgSocket *
-first_socket(StatList *slist)
+first_socket(struct StatList *slist)
 {
 	if (statlist_empty(slist))
 		return NULL;
 	return container_of(slist->head.next, PgSocket, head);
 }
 
-void load_config(bool reload);
+void load_config(void);
 
 
+bool set_config_param(const char *key, const char *val);
+void config_for_each(void (*param_cb)(void *arg, const char *name, const char *val, bool reloadable),
+		     void *arg);
+
