@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -16,17 +16,18 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-extern StatList user_list;
-extern Tree user_tree;
-extern StatList pool_list;
-extern StatList database_list;
-extern StatList login_client_list;
-extern ObjectCache *client_cache;
-extern ObjectCache *server_cache;
-extern ObjectCache *db_cache;
-extern ObjectCache *pool_cache;
-extern ObjectCache *user_cache;
-extern ObjectCache *iobuf_cache;
+extern struct StatList user_list;
+extern struct AATree user_tree;
+extern struct StatList pool_list;
+extern struct StatList database_list;
+extern struct StatList autodatabase_idle_list;
+extern struct StatList login_client_list;
+extern struct Slab *client_cache;
+extern struct Slab *server_cache;
+extern struct Slab *db_cache;
+extern struct Slab *pool_cache;
+extern struct Slab *user_cache;
+extern struct Slab *iobuf_cache;
 
 PgDatabase *find_database(const char *name);
 PgUser *find_user(const char *name);
@@ -34,12 +35,14 @@ PgPool *get_pool(PgDatabase *, PgUser *);
 bool find_server(PgSocket *client)		_MUSTCHECK;
 bool release_server(PgSocket *server)		/* _MUSTCHECK */;
 bool finish_client_login(PgSocket *client)	_MUSTCHECK;
+bool check_fast_fail(PgSocket *client)		_MUSTCHECK;
 
-PgSocket * accept_client(int sock, const struct sockaddr_in *addr, bool is_unix) _MUSTCHECK;
-void disconnect_server(PgSocket *server, bool notify, const char *reason);
-void disconnect_client(PgSocket *client, bool notify, const char *reason);
+PgSocket *accept_client(int sock, bool is_unix) _MUSTCHECK;
+void disconnect_server(PgSocket *server, bool notify, const char *reason, ...) _PRINTF(3, 4);
+void disconnect_client(PgSocket *client, bool notify, const char *reason, ...) _PRINTF(3, 4);
 
 PgDatabase * add_database(const char *name) _MUSTCHECK;
+PgDatabase *register_auto_database(const char *name);
 PgUser * add_user(const char *name, const char *passwd) _MUSTCHECK;
 PgUser * force_user(PgDatabase *db, const char *username, const char *passwd) _MUSTCHECK;
 
