@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -22,7 +22,7 @@
 
 #include "bouncer.h"
 
-#include <regex.h>
+#include <usual/regex.h>
 
 /* regex elements */
 #define WS0	"[ \t\n\r]*"
@@ -64,6 +64,15 @@ static regex_t rc_set_str;
 
 static PgPool *admin_pool;
 
+/* only valid during processing */
+static const char *current_query;
+
+static bool syntax_error(PgSocket *admin)
+{
+	return admin_error(admin, "invalid command '%s', use SHOW HELP;",
+			   current_query ? current_query : "<no query>");
+}
+
 static bool exec_cmd(struct cmd_lookup *lookup, PgSocket *admin,
 		     const char *cmd, const char *arg)
 {
@@ -71,7 +80,7 @@ static bool exec_cmd(struct cmd_lookup *lookup, PgSocket *admin,
 		if (strcasecmp(lookup->word, cmd) == 0)
 			return lookup->func(admin, arg);
 	}
-	return admin_error(admin, "syntax error, use SHOW HELP");
+	return syntax_error(admin);
 }
 
 bool admin_error(PgSocket *admin, const char *fmt, ...)
@@ -92,7 +101,7 @@ bool admin_error(PgSocket *admin, const char *fmt, ...)
 
 static int count_paused_databases(void)
 {
-	List *item;
+	struct List *item;
 	PgDatabase *db;
 	int cnt = 0;
 
@@ -105,7 +114,7 @@ static int count_paused_databases(void)
 
 static int count_db_active(PgDatabase *db)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 	int cnt = 0;
 
@@ -147,6 +156,7 @@ struct FakeParam {
 static const struct FakeParam fake_param_list[] = {
 	{ "client_encoding", "UTF-8" },
 	{ "default_transaction_isolation", "read committed" },
+	{ "standard_conforming_strings", "on" },
 	{ "datestyle", "ISO" },
 	{ "timezone", "GMT" },
 	{ NULL },
@@ -206,12 +216,14 @@ static bool fake_set(PgSocket *admin, const char *key, const char *val)
 static bool admin_set(PgSocket *admin, const char *key, const char *val)
 {
 	char tmp[512];
+	bool ok;
 
 	if (fake_set(admin, key, val))
 		return true;
 
 	if (admin->admin_user) {
-		if (set_config_param(bouncer_params, key, val, true, admin)) {
+		ok = set_config_param(key, val);
+		if (ok) {
 			snprintf(tmp, sizeof(tmp), "SET %s=%s", key, val);
 			return admin_ready(admin, tmp);
 		} else {
@@ -234,18 +246,19 @@ static bool send_one_fd(PgSocket *admin,
 {
 	struct msghdr msg;
 	struct cmsghdr *cmsg;
-	int res;
 	struct iovec iovec;
-	uint8_t pktbuf[1024];
+	int res;
 	uint8_t cntbuf[CMSG_SPACE(sizeof(int))];
 
-	iovec.iov_base = pktbuf;
-	BUILD_DataRow(res, pktbuf, sizeof(pktbuf), "issssiqissss",
+	struct PktBuf *pkt = pktbuf_temp();
+
+	pktbuf_write_DataRow(pkt, "issssiqissss",
 		      fd, task, user, db, addr, port, ckey, link,
 		      client_enc, std_strings, datestyle, timezone);
-	if (res < 0)
+	if (pkt->failed)
 		return false;
-	iovec.iov_len = res;
+	iovec.iov_base = pkt->buf;
+	iovec.iov_len = pktbuf_written(pkt);
 
 	/* sending fds */
 	memset(&msg, 0, sizeof(msg));
@@ -253,7 +266,7 @@ static bool send_one_fd(PgSocket *admin,
 	msg.msg_iovlen = 1;
 
 	/* attach a fd */
-	if (admin->remote_addr.is_unix && admin->own_user) {
+	if (pga_is_unix(&admin->remote_addr) && admin->own_user) {
 		msg.msg_control = cntbuf;
 		msg.msg_controllen = sizeof(cntbuf);
 
@@ -283,47 +296,52 @@ static bool send_one_fd(PgSocket *admin,
 static bool show_one_fd(PgSocket *admin, PgSocket *sk)
 {
 	PgAddr *addr = &sk->remote_addr;
-	MBuf tmp;
+	struct MBuf tmp;
 	VarCache *v = &sk->vars;
+	uint64_t ckey;
+	const struct PStr *client_encoding = v->var_list[VClientEncoding];
+	const struct PStr *std_strings = v->var_list[VStdStr];
+	const struct PStr *datestyle = v->var_list[VDateStyle];
+	const struct PStr *timezone = v->var_list[VTimeZone];
+	char addrbuf[PGADDR_BUF];
+
+	mbuf_init_fixed_reader(&tmp, sk->cancel_key, 8);
+	if (!mbuf_get_uint64be(&tmp, &ckey))
+		return false;
 
-	mbuf_init(&tmp, sk->cancel_key, 8);
 
 	return send_one_fd(admin, sbuf_socket(&sk->sbuf),
 			   is_server_socket(sk) ? "server" : "client",
 			   sk->auth_user ? sk->auth_user->name : NULL,
 			   sk->pool ? sk->pool->db->name : NULL,
-			   addr->is_unix ? "unix" : inet_ntoa(addr->ip_addr),
-			   addr->port,
-			   mbuf_get_uint64(&tmp),
+			   pga_ntop(addr, addrbuf, sizeof(addrbuf)),
+			   pga_port(addr),
+			   ckey,
 			   sk->link ? sbuf_socket(&sk->link->sbuf) : 0,
-			   v->client_encoding[0] ? v->client_encoding : NULL,
-			   v->std_strings[0] ? v->std_strings : NULL,
-			   v->datestyle[0] ? v->datestyle : NULL,
-			   v->timezone[0] ? v->timezone : NULL);
+			   client_encoding ? client_encoding->str : NULL,
+			   std_strings ? std_strings->str : NULL,
+			   datestyle ? datestyle->str : NULL,
+			   timezone ? timezone->str : NULL);
+}
+
+static bool show_pooler_cb(void *arg, int fd, const PgAddr *a)
+{
+	char buf[PGADDR_BUF];
+
+	return send_one_fd(arg, fd, "pooler", NULL, NULL,
+			   pga_ntop(a, buf, sizeof(buf)), pga_port(a), 0, 0,
+			   NULL, NULL, NULL, NULL);
 }
 
 /* send a row with sendmsg, optionally attaching a fd */
 static bool show_pooler_fds(PgSocket *admin)
 {
-	int fd_net, fd_unix;
-	bool res = true;
-
-	get_pooler_fds(&fd_net, &fd_unix);
-
-	if (fd_net)
-		res = send_one_fd(admin, fd_net, "pooler", NULL, NULL,
-				  cf_listen_addr, cf_listen_port, 0, 0,
-				  NULL, NULL, NULL, NULL);
-	if (fd_unix && res)
-		res = send_one_fd(admin, fd_unix, "pooler", NULL, NULL,
-				  "unix", cf_listen_port, 0, 0,
-				  NULL, NULL, NULL, NULL);
-	return res;
+	return for_each_pooler_fd(show_pooler_cb, admin);
 }
 
-static bool show_fds_from_list(PgSocket *admin, StatList *list)
+static bool show_fds_from_list(PgSocket *admin, struct StatList *list)
 {
-	List *item;
+	struct List *item;
 	PgSocket *sk;
 	bool res = true;
 
@@ -343,7 +361,7 @@ static bool show_fds_from_list(PgSocket *admin, StatList *list)
  */
 static bool admin_show_fds(PgSocket *admin, const char *arg)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 	bool res;
 
@@ -404,8 +422,7 @@ static bool admin_show_fds(PgSocket *admin, const char *arg)
 static bool admin_show_databases(PgSocket *admin, const char *arg)
 {
 	PgDatabase *db;
-	List *item;
-	char *host;
+	struct List *item;
 	const char *f_user;
 	PktBuf *buf;
 
@@ -415,22 +432,18 @@ static bool admin_show_databases(PgSocket *admin, const char *arg)
 		return true;
 	}
 
-	pktbuf_write_RowDescription(buf, "ssissi",
+	pktbuf_write_RowDescription(buf, "ssissii",
 				    "name", "host", "port",
-				    "database", "force_user", "pool_size");
+				    "database", "force_user", "pool_size", "reserve_pool");
 	statlist_for_each(item, &database_list) {
 		db = container_of(item, PgDatabase, head);
 
-		if (!db->addr.is_unix) {
-			host = inet_ntoa(db->addr.ip_addr);
-		} else
-			host = NULL;
-
 		f_user = db->forced_user ? db->forced_user->name : NULL;
-		pktbuf_write_DataRow(buf, "ssissi",
-				     db->name, host, db->addr.port,
+		pktbuf_write_DataRow(buf, "ssissii",
+				     db->name, db->host, db->port,
 				     db->dbname, f_user,
-				     db->pool_size);
+				     db->pool_size,
+				     db->res_pool_size);
 	}
 	admin_flush(admin, buf, "SHOW");
 	return true;
@@ -450,11 +463,11 @@ static bool admin_show_lists(PgSocket *admin, const char *arg)
 	SENDLIST("databases", statlist_count(&database_list));
 	SENDLIST("users", statlist_count(&user_list));
 	SENDLIST("pools", statlist_count(&pool_list));
-	SENDLIST("free_clients", objcache_free_count(client_cache));
-	SENDLIST("used_clients", objcache_active_count(client_cache));
+	SENDLIST("free_clients", slab_free_count(client_cache));
+	SENDLIST("used_clients", slab_active_count(client_cache));
 	SENDLIST("login_clients", statlist_count(&login_client_list));
-	SENDLIST("free_servers", objcache_free_count(server_cache));
-	SENDLIST("used_servers", objcache_active_count(server_cache));
+	SENDLIST("free_servers", slab_free_count(server_cache));
+	SENDLIST("used_servers", slab_active_count(server_cache));
 	admin_flush(admin, buf, "SHOW");
 	return true;
 }
@@ -463,7 +476,7 @@ static bool admin_show_lists(PgSocket *admin, const char *arg)
 static bool admin_show_users(PgSocket *admin, const char *arg)
 {
 	PgUser *user;
-	List *item;
+	struct List *item;
 	PktBuf *buf = pktbuf_dynamic(256);
 	if (!buf) {
 		admin_error(admin, "no mem");
@@ -495,12 +508,7 @@ static void socket_header(PktBuf *buf, bool debug)
 
 static void adr2txt(const PgAddr *adr, char *dst, unsigned dstlen)
 {
-	if (adr->is_unix) {
-		safe_strcpy(dst, "unix", dstlen);
-	} else {
-		char *tmp = inet_ntoa(adr->ip_addr);
-		safe_strcpy(dst, tmp, dstlen);
-	}
+	pga_ntop(adr, dst, dstlen);
 }
 
 static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
@@ -528,8 +536,8 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 			     is_server_socket(sk) ? "S" :"C",
 			     sk->auth_user ? sk->auth_user->name : "(nouser)",
 			     sk->pool ? sk->pool->db->name : "(nodb)",
-			     state, r_addr, sk->remote_addr.port,
-			     l_addr, sk->local_addr.port,
+			     state, r_addr, pga_port(&sk->remote_addr),
+			     l_addr, pga_port(&sk->local_addr),
 			     sk->connect_time,
 			     sk->request_time,
 			     ptrbuf, linkbuf,
@@ -542,9 +550,9 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 }
 
 /* Helper for SHOW CLIENTS */
-static void show_socket_list(PktBuf *buf, StatList *list, const char *state, bool debug)
+static void show_socket_list(PktBuf *buf, struct StatList *list, const char *state, bool debug)
 {
-	List *item;
+	struct List *item;
 	PgSocket *sk;
 
 	statlist_for_each(item, list) {
@@ -556,7 +564,7 @@ static void show_socket_list(PktBuf *buf, StatList *list, const char *state, boo
 /* Command: SHOW CLIENTS */
 static bool admin_show_clients(PgSocket *admin, const char *arg)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 	PktBuf *buf = pktbuf_dynamic(256);
 
@@ -580,7 +588,7 @@ static bool admin_show_clients(PgSocket *admin, const char *arg)
 /* Command: SHOW SERVERS */
 static bool admin_show_servers(PgSocket *admin, const char *arg)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 	PktBuf *buf;
 
@@ -606,7 +614,7 @@ static bool admin_show_servers(PgSocket *admin, const char *arg)
 /* Command: SHOW SOCKETS */
 static bool admin_show_sockets(PgSocket *admin, const char *arg)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 	PktBuf *buf;
 
@@ -633,9 +641,9 @@ static bool admin_show_sockets(PgSocket *admin, const char *arg)
 	return true;
 }
 
-static void show_active_socket_list(PktBuf *buf, StatList *list, const char *state)
+static void show_active_socket_list(PktBuf *buf, struct StatList *list, const char *state)
 {
-	List *item;
+	struct List *item;
 	statlist_for_each(item, list) {
 		PgSocket *sk = container_of(item, PgSocket, head);
 		if (!sbuf_is_empty(&sk->sbuf))
@@ -646,7 +654,7 @@ static void show_active_socket_list(PktBuf *buf, StatList *list, const char *sta
 /* Command: SHOW ACTIVE_SOCKETS */
 static bool admin_show_active_sockets(PgSocket *admin, const char *arg)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 	PktBuf *buf;
 
@@ -676,7 +684,7 @@ static bool admin_show_active_sockets(PgSocket *admin, const char *arg)
 /* Command: SHOW POOLS */
 static bool admin_show_pools(PgSocket *admin, const char *arg)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 	PktBuf *buf;
 	PgSocket *waiter;
@@ -735,16 +743,21 @@ static bool admin_show_mem(PgSocket *admin, const char *arg)
 	}
 	pktbuf_write_RowDescription(buf, "siiii", "name",
 				    "size", "used", "free", "memtotal");
-	objcache_stats(slab_stat_cb, buf);
+	slab_stats(slab_stat_cb, buf);
 	admin_flush(admin, buf, "SHOW");
 	return true;
 }
 
+static void show_one_param(void *arg, const char *name, const char *val, bool reloadable)
+{
+	PktBuf *buf = arg;
+	pktbuf_write_DataRow(buf, "sss", name, val,
+			     reloadable ? "yes" : "no");
+}
+
 /* Command: SHOW CONFIG */
 static bool admin_show_config(PgSocket *admin, const char *arg)
 {
-	ConfElem *cf;
-	int i = 0;
 	PktBuf *buf;
 
 	buf = pktbuf_dynamic(256);
@@ -754,16 +767,11 @@ static bool admin_show_config(PgSocket *admin, const char *arg)
 	}
 
 	pktbuf_write_RowDescription(buf, "sss", "key", "value", "changeable");
-	while (1) {
-		cf = &bouncer_params[i++];
-		if (!cf->name)
-			break;
 
-		pktbuf_write_DataRow(buf, "sss",
-				     cf->name, conf_to_text(cf),
-				     cf->reloadable ? "yes" : "no");
-	}
+	config_for_each(show_one_param, buf);
+
 	admin_flush(admin, buf, "SHOW");
+
 	return true;
 }
 
@@ -771,13 +779,13 @@ static bool admin_show_config(PgSocket *admin, const char *arg)
 static bool admin_cmd_reload(PgSocket *admin, const char *arg)
 {
 	if (arg && *arg)
-		return admin_error(admin, "syntax error");
+		return syntax_error(admin);
 
 	if (!admin->admin_user)
 		return admin_error(admin, "admin access needed");
 
 	log_info("RELOAD command issued");
-	load_config(true);
+	load_config();
 	return admin_ready(admin, "RELOAD");
 }
 
@@ -785,7 +793,7 @@ static bool admin_cmd_reload(PgSocket *admin, const char *arg)
 static bool admin_cmd_shutdown(PgSocket *admin, const char *arg)
 {
 	if (arg && *arg)
-		return admin_error(admin, "syntax error");
+		return syntax_error(admin);
 
 	if (!admin->admin_user)
 		return admin_error(admin, "admin access needed");
@@ -796,7 +804,8 @@ static bool admin_cmd_shutdown(PgSocket *admin, const char *arg)
 	 * before closing open sockets.
 	 */
 	log_info("SHUTDOWN command issued");
-	exit(0);
+	cf_shutdown = 2;
+	event_loopbreak();
 
 	return true;
 }
@@ -843,7 +852,7 @@ static bool admin_cmd_resume(PgSocket *admin, const char *arg)
 static bool admin_cmd_suspend(PgSocket *admin, const char *arg)
 {
 	if (arg && *arg)
-		return admin_error(admin, "syntax error");
+		return syntax_error(admin);
 
 	if (!admin->admin_user)
 		return admin_error(admin, "admin access needed");
@@ -919,7 +928,6 @@ static void copy_arg_unquote(const char *str, regmatch_t *glist,
 	const char *end = src + len;
 
 	if (len < dstmax) {
-		len = 0;
 		while (src < end) {
 			if (src[0] == '\'' && src[1] == '\'') {
 				*dst++ = '\'';
@@ -939,7 +947,7 @@ static bool admin_show_help(PgSocket *admin, const char *arg)
 		"SNOTICE", "C00000", "MConsole usage",
 		"D\n\tSHOW HELP|CONFIG|DATABASES"
 		"|POOLS|CLIENTS|SERVERS|VERSION\n"
-		"\tSHOW FDS|SOCKETS|ACTIVE_SOCKETS|LISTS|MEM\n"
+		"\tSHOW STATS|FDS|SOCKETS|ACTIVE_SOCKETS|LISTS|MEM\n"
 		"\tSET key = arg\n"
 		"\tRELOAD\n"
 		"\tPAUSE [<db>]\n"
@@ -1019,6 +1027,8 @@ static bool admin_parse_query(PgSocket *admin, const char *q)
 	char val[256];
 	bool res;
 
+	current_query = q;
+
 	if (regexec(&rc_cmd, q, MAX_GROUPS, grp, 0) == 0) {
 		copy_arg(q, grp, CMD_NAME, cmd, sizeof(cmd));
 		copy_arg(q, grp, CMD_ARG, arg, sizeof(arg));
@@ -1038,7 +1048,9 @@ static bool admin_parse_query(PgSocket *admin, const char *q)
 		} else
 			res = admin_set(admin, arg, val);
 	} else
-		res = admin_error(admin, "unknown cmd: %s", q);
+		res = syntax_error(admin);
+
+	current_query = NULL;
 
 	if (!res)
 		disconnect_client(admin, true, "failure");
@@ -1059,8 +1071,7 @@ bool admin_handle_client(PgSocket *admin, PktHdr *pkt)
 
 	switch (pkt->type) {
 	case 'Q':
-		q = mbuf_get_string(&pkt->data);
-		if (!q) {
+		if (!mbuf_get_string(&pkt->data, &q)) {
 			disconnect_client(admin, true, "incomplete query");
 			return false;
 		}
@@ -1095,7 +1106,7 @@ bool admin_pre_login(PgSocket *client)
 	client->own_user = 0;
 
 	/* tag same uid as special */
-	if (client->remote_addr.is_unix) {
+	if (pga_is_unix(&client->remote_addr)) {
 		res = getpeereid(sbuf_socket(&client->sbuf), &peer_uid, &peer_gid);
 		if (res >= 0 && peer_uid == getuid()
 			&& strcmp("pgbouncer", username) == 0)
@@ -1107,6 +1118,17 @@ bool admin_pre_login(PgSocket *client)
 		}
 	}
 
+	/*
+	 * auth_mode=any does not keep original username around,
+	 * so username based checks do not work.
+	 */
+	if (cf_auth_type == AUTH_ANY) {
+		if (cf_log_connections)
+			slog_info(client, "auth_mode=any: allowing anybody in as admin");
+		client->admin_user = 1;
+		return true;
+	}
+
 	if (strlist_contains(cf_admin_users, username)) {
 		client->admin_user = 1;
 		return true;
@@ -1123,7 +1145,7 @@ void admin_setup(void)
 	PgDatabase *db;
 	PgPool *pool;
 	PgUser *user;
-	PktBuf msg;
+	PktBuf *msg;
 	int res;
 
 	/* fake database */
@@ -1131,8 +1153,7 @@ void admin_setup(void)
 	if (!db)
 		fatal("no memory for admin database");
 
-	db->addr.port = cf_listen_port;
-	db->addr.is_unix = 1;
+	db->port = cf_listen_port;
 	db->pool_size = 2;
 	db->admin = 1;
 	if (!force_user(db, "pgbouncer", ""))
@@ -1150,21 +1171,31 @@ void admin_setup(void)
 		fatal("cannot create admin user?");
 
 	/* prepare welcome */
-	pktbuf_static(&msg, pool->welcome_msg, sizeof(pool->welcome_msg));
-	pktbuf_write_AuthenticationOk(&msg);
-	pktbuf_write_ParameterStatus(&msg, "server_version", "8.0/bouncer");
-	pktbuf_write_ParameterStatus(&msg, "client_encoding", "UNICODE");
-	pktbuf_write_ParameterStatus(&msg, "server_encoding", "SQL_ASCII");
-	pktbuf_write_ParameterStatus(&msg, "is_superuser", "on");
+	msg = pktbuf_dynamic(128);
+	if (!msg)
+		fatal("cannot create admin welcome");
+	pktbuf_write_AuthenticationOk(msg);
+	pktbuf_write_ParameterStatus(msg, "server_version", PACKAGE_VERSION "/bouncer");
+	pktbuf_write_ParameterStatus(msg, "client_encoding", "UNICODE");
+	pktbuf_write_ParameterStatus(msg, "server_encoding", "SQL_ASCII");
+	pktbuf_write_ParameterStatus(msg, "DateStyle", "ISO");
+	pktbuf_write_ParameterStatus(msg, "TimeZone", "GMT");
+	pktbuf_write_ParameterStatus(msg, "standard_conforming_strings", "on");
+	pktbuf_write_ParameterStatus(msg, "is_superuser", "on");
 
-	pool->welcome_msg_len = pktbuf_written(&msg);
+	if (msg->failed)
+		fatal("admin welcome failed");
+
+	pool->welcome_msg = msg;
 	pool->welcome_msg_ready = 1;
 
-	pktbuf_static(&msg, db->startup_params, sizeof(db->startup_params));
-	pktbuf_put_string(&msg, "database");
-	db->dbname = (char *)db->startup_params + pktbuf_written(&msg);
-	pktbuf_put_string(&msg, "pgbouncer");
-	db->startup_params_len = pktbuf_written(&msg);
+	msg = pktbuf_dynamic(128);
+	if (!msg)
+		fatal("cannot create admin startup pkt");
+	db->startup_params = msg;
+	pktbuf_put_string(msg, "database");
+	db->dbname = "pgbouncer";
+	pktbuf_put_string(msg, db->dbname);
 
 	/* initialize regexes */
 	res = regcomp(&rc_cmd, cmd_normal_rx, REG_EXTENDED | REG_ICASE);
@@ -1180,7 +1211,7 @@ void admin_setup(void)
 
 void admin_pause_done(void)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgSocket *admin;
 	bool res;
 
