@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -37,7 +37,7 @@ static bool check_client_passwd(PgSocket *client, const char *passwd)
 		return strcmp(user->passwd, passwd) == 0;
 	case AUTH_CRYPT:
 		correct = crypt(user->passwd, (char *)client->tmp_login_salt);
-		return strcmp(correct, passwd) == 0;
+		return correct && strcmp(correct, passwd) == 0;
 	case AUTH_MD5:
 		if (strlen(passwd) != MD5_PASSWD_LEN)
 			return false;
@@ -57,8 +57,14 @@ bool set_pool(PgSocket *client, const char *dbname, const char *username)
 	/* find database */
 	db = find_database(dbname);
 	if (!db) {
-		disconnect_client(client, true, "No such database");
-		return false;
+		db = register_auto_database(dbname);
+		if (!db) {
+			disconnect_client(client, true, "No such database: %s", dbname);
+			return false;
+		}
+		else {
+			slog_info(client, "registered new auto-database: db = %s", dbname );
+		}
 	}
 
 	/* find user */
@@ -76,7 +82,7 @@ bool set_pool(PgSocket *client, const char *dbname, const char *username)
 		/* the user clients wants to log in as */
 		user = find_user(username);
 		if (!user) {
-			disconnect_client(client, true, "No such user");
+			disconnect_client(client, true, "No such user: %s", username);
 			return false;
 		}
 		client->auth_user = user;
@@ -91,44 +97,47 @@ bool set_pool(PgSocket *client, const char *dbname, const char *username)
 		return false;
 	}
 
-	return true;
+	return check_fast_fail(client);
 }
 
 static bool decide_startup_pool(PgSocket *client, PktHdr *pkt)
 {
 	const char *username = NULL, *dbname = NULL;
 	const char *key, *val;
+	bool ok;
 
 	while (1) {
-		key = mbuf_get_string(&pkt->data);
-		if (!key || *key == 0)
+		ok = mbuf_get_string(&pkt->data, &key);
+		if (!ok || *key == 0)
 			break;
-		val = mbuf_get_string(&pkt->data);
-		if (!val)
+		ok = mbuf_get_string(&pkt->data, &val);
+		if (!ok)
 			break;
 
-		if (strcmp(key, "database") == 0)
-			dbname = val;
-		else if (strcmp(key, "user") == 0)
-			username = val;
-		else if (varcache_set(&client->vars, key, val))
+		if (strcmp(key, "database") == 0) {
 			slog_debug(client, "got var: %s=%s", key, val);
-		else if (strlist_contains(cf_ignore_startup_params, key)) {
+			dbname = val;
+		} else if (strcmp(key, "user") == 0) {
+			slog_debug(client, "got var: %s=%s", key, val);
+			username = val;
+		} else if (varcache_set(&client->vars, key, val)) {
+			slog_debug(client, "got var: %s=%s", key, val);
+		} else if (strlist_contains(cf_ignore_startup_params, key)) {
 			slog_debug(client, "ignoring startup parameter: %s=%s", key, val);
 		} else {
 			slog_warning(client, "unsupported startup parameter: %s=%s", key, val);
-			disconnect_client(client, true, "Unknown startup parameter");
+			disconnect_client(client, true, "Unsupported startup parameter: %s", key);
 			return false;
 		}
 	}
-	if (!username) {
+	if (!username || !username[0]) {
 		disconnect_client(client, true, "No username supplied");
 		return false;
 	}
-	if (!dbname) {
-		disconnect_client(client, true, "No database supplied");
-		return false;
-	}
+
+	/* if missing dbname, default to username */
+	if (!dbname || !dbname[0])
+		dbname = username;
 
 	/* check if limit allows, dont limit admin db
 	   nb: new incoming conn will be attached to PgSocket, thus
@@ -143,7 +152,7 @@ static bool decide_startup_pool(PgSocket *client, PktHdr *pkt)
 	/* find pool and log about it */
 	if (set_pool(client, dbname, username)) {
 		if (cf_log_connections)
-			slog_info(client, "login successful: db=%s user=%s", dbname, username);
+			slog_info(client, "login attempt: db=%s user=%s", dbname, username);
 		return true;
 	} else {
 		if (cf_log_connections)
@@ -185,6 +194,8 @@ static bool send_client_authreq(PgSocket *client)
 static bool handle_client_startup(PgSocket *client, PktHdr *pkt)
 {
 	const char *passwd;
+	const uint8_t *key;
+	bool ok;
 
 	SBuf *sbuf = &client->sbuf;
 
@@ -214,6 +225,9 @@ static bool handle_client_startup(PgSocket *client, PktHdr *pkt)
 			return false;
 		}
 		break;
+	case PKT_STARTUP_V2:
+		disconnect_client(client, true, "Old V2 protocol not supported");
+		return false;
 	case PKT_STARTUP:
 		if (client->pool) {
 			disconnect_client(client, true, "client re-sent startup pkt");
@@ -245,18 +259,19 @@ static bool handle_client_startup(PgSocket *client, PktHdr *pkt)
 			return false;
 		}
 
-		passwd = mbuf_get_string(&pkt->data);
-		if (passwd && check_client_passwd(client, passwd)) {
+		ok = mbuf_get_string(&pkt->data, &passwd);
+		if (ok && check_client_passwd(client, passwd)) {
 			if (!finish_client_login(client))
 				return false;
 		} else {
-			disconnect_client(client, true, "Login failed");
+			disconnect_client(client, true, "Auth failed");
 			return false;
 		}
 		break;
 	case PKT_CANCEL:
-		if (mbuf_avail(&pkt->data) == BACKENDKEY_LEN) {
-			const uint8_t *key = mbuf_get_bytes(&pkt->data, BACKENDKEY_LEN);
+		if (mbuf_avail_for_read(&pkt->data) == BACKENDKEY_LEN
+		    && mbuf_get_bytes(&pkt->data, BACKENDKEY_LEN, &key))
+		{
 			memcpy(client->cancel_key, key, BACKENDKEY_LEN);
 			accept_cancel_request(client);
 		} else
@@ -278,14 +293,19 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 
 	switch (pkt->type) {
 
+	/* one-packet queries */
+	case 'Q':		/* Query */
+		if (cf_disable_pqexec) {
+			slog_error(client, "Client used 'Q' packet type.");
+			disconnect_client(client, true, "PQexec disallowed");
+			return false;
+		}
+	case 'F':		/* FunctionCall */
+
 	/* request immidiate response from server */
 	case 'H':		/* Flush */
 	case 'S':		/* Sync */
 
-	/* one-packet queries */
-	case 'Q':		/* Query */
-	case 'F':		/* FunctionCall */
-
 	/* copy end markers */
 	case 'c':		/* CopyDone(F/B) */
 	case 'f':		/* CopyFail(F/B) */
@@ -336,7 +356,7 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 }
 
 /* callback from SBuf */
-bool client_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data)
+bool client_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 {
 	bool res = false;
 	PgSocket *client = container_of(sbuf, PgSocket, sbuf);
@@ -362,7 +382,7 @@ bool client_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data)
 		disconnect_server(client->link, false, "Server connection closed");
 		break;
 	case SBUF_EV_READ:
-		if (mbuf_avail(data) < NEW_HEADER_LEN && client->state != CL_LOGIN) {
+		if (mbuf_avail_for_read(data) < NEW_HEADER_LEN && client->state != CL_LOGIN) {
 			slog_noise(client, "C: got partial header, trying to wait a bit");
 			return false;
 		}
