@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -27,25 +27,25 @@ static struct timeval full_maint_period = {0, USEC / 3};
 static struct event full_maint_ev;
 
 /* close all sockets in server list */
-static void close_server_list(StatList *sk_list, const char *reason)
+static void close_server_list(struct StatList *sk_list, const char *reason)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgSocket *server;
 
 	statlist_for_each_safe(item, sk_list, tmp) {
 		server = container_of(item, PgSocket, head);
-		disconnect_server(server, true, reason);
+		disconnect_server(server, true, "%s", reason);
 	}
 }
 
-static void close_client_list(StatList *sk_list, const char *reason)
+static void close_client_list(struct StatList *sk_list, const char *reason)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgSocket *client;
 
 	statlist_for_each_safe(item, sk_list, tmp) {
 		client = container_of(item, PgSocket, head);
-		disconnect_client(client, true, reason);
+		disconnect_client(client, true, "%s", reason);
 	}
 }
 
@@ -70,9 +70,9 @@ bool suspend_socket(PgSocket *sk, bool force_suspend)
 }
 
 /* suspend all sockets in socket list */
-static int suspend_socket_list(StatList *list, bool force_suspend)
+static int suspend_socket_list(struct StatList *list, bool force_suspend)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgSocket *sk;
 	int active = 0;
 
@@ -85,9 +85,9 @@ static int suspend_socket_list(StatList *list, bool force_suspend)
 }
 
 /* resume all suspended sockets in socket list */
-static void resume_socket_list(StatList *list)
+static void resume_socket_list(struct StatList *list)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgSocket *sk;
 
 	statlist_for_each_safe(item, list, tmp) {
@@ -102,7 +102,7 @@ static void resume_socket_list(StatList *list)
 /* resume all suspended sockets in all pools */
 static void resume_sockets(void)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 
 	statlist_for_each(item, &pool_list) {
@@ -169,7 +169,7 @@ static void launch_recheck(PgPool *pool)
  */
 static void per_loop_activate(PgPool *pool)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgSocket *client;
 
 	/* see if any server have been freed */
@@ -254,7 +254,7 @@ static int per_loop_suspend(PgPool *pool, bool force_suspend)
  */
 void per_loop_maint(void)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 	int active = 0;
 	int partial_pause = 0;
@@ -307,7 +307,7 @@ void per_loop_maint(void)
 /* maintaining clients in pool */
 static void pool_client_maint(PgPool *pool)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	usec_t now = get_cached_time();
 	PgSocket *client;
 	usec_t age;
@@ -324,8 +324,8 @@ static void pool_client_maint(PgPool *pool)
 		}
 	}
 
-	/* force client_query_timeout */
-	if (cf_query_timeout > 0) {
+	/* force timeouts for waiting queries */
+	if (cf_query_timeout > 0 || cf_query_wait_timeout > 0) {
 		statlist_for_each_safe(item, &pool->waiting_client_list, tmp) {
 			client = container_of(item, PgSocket, head);
 			Assert(client->state == CL_WAITING);
@@ -334,16 +334,31 @@ static void pool_client_maint(PgPool *pool)
 				//log_warning("query_start==0");
 			} else
 				age = now - client->query_start;
-			if (age > cf_query_timeout)
+
+			if (cf_query_timeout > 0 && age > cf_query_timeout)
 				disconnect_client(client, true, "query_timeout");
+			else if (cf_query_wait_timeout > 0 && age > cf_query_wait_timeout)
+				disconnect_client(client, true, "query_wait_timeout");
+		}
+	}
+
+	/* apply client_login_timeout to clients waiting for welcome pkt */
+	if (cf_client_login_timeout > 0 && !pool->welcome_msg_ready) {
+		statlist_for_each_safe(item, &pool->waiting_client_list, tmp) {
+			client = container_of(item, PgSocket, head);
+			if (!client->wait_for_welcome)
+				continue;
+			age = now - client->connect_time;
+			if (age > cf_client_login_timeout)
+				disconnect_client(client, true, "client_login_timeout (server down)");
 		}
 	}
 }
 
-static void check_unused_servers(PgPool *pool, StatList *slist, bool idle_test)
+static void check_unused_servers(PgPool *pool, struct StatList *slist, bool idle_test)
 {
 	usec_t now = get_cached_time();
-	List *item, *tmp;
+	struct List *item, *tmp;
 	usec_t idle, age;
 	PgSocket *server;
 	usec_t lifetime_kill_gap = 0;
@@ -403,7 +418,7 @@ static void check_pool_size(PgPool *pool)
 		 * statlist_count(&pool->new_server_list)
 		 */
 
-	int many = cur - pool->db->pool_size;
+	int many = cur - (pool->db->pool_size + pool->db->res_pool_size);
 
 	Assert(pool->db->pool_size >= 0);
 
@@ -421,7 +436,7 @@ static void check_pool_size(PgPool *pool)
 /* maintain servers in a pool */
 static void pool_server_maint(PgPool *pool)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	usec_t age, now = get_cached_time();
 	PgSocket *server;
 
@@ -439,7 +454,7 @@ static void pool_server_maint(PgPool *pool)
 				continue;
 			age = now - server->link->request_time;
 			if (age > cf_query_timeout)
-				disconnect_server(server, true, "statement timeout");
+				disconnect_server(server, true, "query timeout");
 		}
 	}
 
@@ -460,7 +475,7 @@ static void pool_server_maint(PgPool *pool)
 
 static void cleanup_client_logins(void)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgSocket *client;
 	usec_t age;
 	usec_t now = get_cached_time();
@@ -476,30 +491,85 @@ static void cleanup_client_logins(void)
 	}
 }
 
+static void kill_database(PgDatabase *db);
+static void cleanup_inactive_autodatabases(void)
+{
+	struct List *item, *tmp;
+	PgDatabase *db;
+	usec_t age;
+	usec_t now = get_cached_time();
+
+	if (cf_autodb_idle_timeout <= 0)
+		return;
+
+	/* now kill the old ones */
+	statlist_for_each_safe(item, &autodatabase_idle_list, tmp) {
+		db = container_of(item, PgDatabase, head);
+		age = now - db->inactive_time;
+		if (age > cf_autodb_idle_timeout) 
+			kill_database(db);
+		else
+			break;
+	}
+}
+
 /* full-scale maintenance, done only occasionally */
 static void do_full_maint(int sock, short flags, void *arg)
 {
-	List *item;
+	struct List *item, *tmp;
 	PgPool *pool;
+	PgDatabase *db;
 
-	statlist_for_each(item, &pool_list) {
+	static unsigned int seq;
+	seq++;
+
+	/*
+	 * Avoid doing anything that may surprise other pgbouncer.
+	 */
+	if (cf_pause_mode == P_SUSPEND)
+		goto skip_maint;
+
+	statlist_for_each_safe(item, &pool_list, tmp) {
 		pool = container_of(item, PgPool, head);
 		if (pool->db->admin)
 			continue;
 		pool_server_maint(pool);
 		pool_client_maint(pool);
+
+		/* is autodb active? */
+		if (pool->db->db_auto && pool->db->inactive_time == 0) {
+			if (pool_client_count(pool) > 0 || pool_server_count(pool) > 0)
+				pool->db->active_stamp = seq;
+		}
 	}
 
+	/* find inactive autodbs */
+	statlist_for_each_safe(item, &database_list, tmp) {
+		db = container_of(item, PgDatabase, head);
+		if (db->db_auto && db->inactive_time == 0) {
+			if (db->active_stamp == seq)
+				continue;
+			db->inactive_time = get_cached_time();
+			statlist_remove(&database_list, &db->head);
+			statlist_append(&autodatabase_idle_list, &db->head);
+		}
+	}
+
+	cleanup_inactive_autodatabases();
+
 	cleanup_client_logins();
 
-	if (cf_shutdown && get_active_server_count() == 0) {
+	if (cf_shutdown == 1 && get_active_server_count() == 0) {
 		log_info("server connections dropped, exiting");
-		exit(0);
+		cf_shutdown = 2;
+		event_loopbreak();
+		return;
 	}
 
 	if (cf_auth_type >= AUTH_TRUST)
 		loader_users_check();
 
+skip_maint:
 	safe_evtimer_add(&full_maint_ev, &full_maint_period);
 }
 
@@ -525,44 +595,55 @@ static void kill_pool(PgPool *pool)
 	close_server_list(&pool->tested_server_list, reason);
 	close_server_list(&pool->new_server_list, reason);
 
+	pktbuf_free(pool->welcome_msg);
+
 	list_del(&pool->map_head);
-	statlist_remove(&pool->head, &pool_list);
-	obj_free(pool_cache, pool);
+	statlist_remove(&pool_list, &pool->head);
+	varcache_clean(&pool->orig_vars);
+	slab_free(pool_cache, pool);
 }
 
 static void kill_database(PgDatabase *db)
 {
 	PgPool *pool;
-	List *item, *tmp;
+	struct List *item, *tmp;
 
-	log_warning("dropping database '%s' as it does not exist anymore", db->name);
+	log_warning("dropping database '%s' as it does not exist anymore or inactive auto-database", db->name);
 
 	statlist_for_each_safe(item, &pool_list, tmp) {
 		pool = container_of(item, PgPool, head);
 		if (pool->db == db)
 			kill_pool(pool);
 	}
+	pktbuf_free(db->startup_params);
 	if (db->forced_user)
-		obj_free(user_cache, db->forced_user);
+		slab_free(user_cache, db->forced_user);
 	if (db->connect_query)
 		free((void *)db->connect_query);
-	statlist_remove(&db->head, &database_list);
-	obj_free(db_cache, db);
+	if (db->inactive_time)
+		statlist_remove(&autodatabase_idle_list, &db->head);
+	else
+		statlist_remove(&database_list, &db->head);
+	slab_free(db_cache, db);
 }
 
 /* as [pgbouncer] section can be loaded after databases,
    there's need for review */
 void config_postprocess(void)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgDatabase *db;
 
 	statlist_for_each_safe(item, &database_list, tmp) {
 		db = container_of(item, PgDatabase, head);
-		if (db->db_dead)
+		if (db->db_dead) {
 			kill_database(db);
-		else if (db->pool_size < 0)
+			continue;
+		}
+		if (db->pool_size < 0)
 			db->pool_size = cf_default_pool_size;
+		if (db->res_pool_size < 0)
+			db->res_pool_size = cf_res_pool_size;
 	}
 }
 
