@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -22,7 +22,11 @@
 
 #include "bouncer.h"
 
+#ifdef HAVE_NETDB_H
 #include <netdb.h>
+#endif
+
+#include <usual/fileutil.h>
 
 /*
  * ConnString parsing
@@ -152,15 +156,30 @@ static void set_connect_query(PgDatabase *db, const char *new)
 		log_error("no memory, cannot assign connect_query for %s", db->name);
 }
 
+static void set_autodb(const char *connstr)
+{
+	char *tmp = strdup(connstr);
+	if (!tmp) {
+		log_warning("no mem to change autodb_connstr");
+		return;
+	}
+	if (cf_autodb_connstr)
+		free(cf_autodb_connstr);
+	cf_autodb_connstr = tmp;
+}
+
 /* fill PgDatabase from connstr */
-void parse_database(char *name, char *connstr)
+bool parse_database(void *base, const char *name, const char *connstr)
 {
 	char *p, *key, *val;
-	PktBuf buf;
+	PktBuf *msg;
 	PgDatabase *db;
 	int pool_size = -1;
+	int res_pool_size = -1;
+	int dbname_ofs;
 
-	char *dbname = name;
+	char *tmp_connstr;
+	const char *dbname = name;
 	char *host = NULL;
 	char *port = "5432";
 	char *username = NULL;
@@ -169,17 +188,25 @@ void parse_database(char *name, char *connstr)
 	char *datestyle = NULL;
 	char *timezone = NULL;
 	char *connect_query = NULL;
-	char *unix_dir = "";
+	char *appname = NULL;
 
-	in_addr_t v_addr = INADDR_NONE;
 	int v_port;
 
-	p = connstr;
+	if (strcmp(name, "*") == 0) {
+		set_autodb(connstr);
+		return true;
+	}
+
+	tmp_connstr = strdup(connstr);
+	if (!tmp_connstr)
+		return false;
+
+	p = tmp_connstr;
 	while (*p) {
 		p = cstr_get_pair(p, &key, &val);
 		if (p == NULL) {
 			log_error("%s: syntax error in connstring", name);
-			return;
+			goto fail;
 		} else if (!key[0])
 			break;
 
@@ -201,82 +228,58 @@ void parse_database(char *name, char *connstr)
 			timezone = val;
 		else if (strcmp("pool_size", key) == 0)
 			pool_size = atoi(val);
+		else if (strcmp("reserve_pool", key) == 0)
+			res_pool_size = atoi(val);
 		else if (strcmp("connect_query", key) == 0)
 			connect_query = val;
+		else if (strcmp("application_name", key) == 0)
+			appname = val;
 		else {
 			log_error("skipping database %s because"
 				  " of unknown parameter in connstring: %s", name, key);
-			return;
+			goto fail;
 		}
 	}
 
-	/* host= */
-	if (!host) {
-		/* default unix socket dir */
-		if (!cf_unix_socket_dir) {
-			log_error("skipping database %s because"
-				" unix socket not configured", name);
-			return;
-		}
-	} else if (host[0] == '/') {
-		/* custom unix socket dir */
-		unix_dir = host;
-		host = NULL;
-	} else if (host[0] >= '0' && host[0] <= '9') {
-		/* ip-address */
-		v_addr = inet_addr(host);
-		if (v_addr == INADDR_NONE) {
-			log_error("skipping database %s because"
-					" of bad host: %s", name, host);
-			return;
-		}
-	} else {
-		/* resolve host by name */
-		struct hostent *h = gethostbyname(host);
-		if (h == NULL || h->h_addr_list[0] == NULL) {
-			log_error("%s: resolving host=%s failed: %s",
-				  name, host, hstrerror(h_errno));
-			return;
-		}
-		if (h->h_addrtype != AF_INET || h->h_length != 4) {
-			log_error("%s: host=%s has unknown addr type",
-				  name, host);
-			return;
-		}
-
-		/* result should be already in correct endianess */
-		memcpy(&v_addr, h->h_addr_list[0], 4);
-	}
-
 	/* port= */
 	v_port = atoi(port);
 	if (v_port == 0) {
 		log_error("skipping database %s because"
 			  " of bad port: %s", name, port);
-		return;
+		goto fail;
 	}
 
 	db = add_database(name);
 	if (!db) {
 		log_error("cannot create database, no memory?");
-		return;
+		goto fail;
+	}
+
+	/* host= */
+	if (host) {
+		host = strdup(host);
+		if (!host) {
+			log_error("failed to allocate host=");
+			goto fail;
+		}
 	}
 
 	/* tag the db as alive */
 	db->db_dead = 0;
+	/* assuming not an autodb */
+	db->db_auto = 0;
+	db->inactive_time = 0;
 
 	/* if updating old db, check if anything changed */
 	if (db->dbname) {
 		bool changed = false;
 		if (strcmp(db->dbname, dbname) != 0)
 			changed = true;
-		else if (host && db->addr.is_unix)
+		else if (!!host != !!db->host)
 			changed = true;
-		else if (!host && !db->addr.is_unix)
+		else if (host && strcmp(host, db->host) != 0)
 			changed = true;
-		else if (host && v_addr != db->addr.ip_addr.s_addr)
-			changed = true;
-		else if (v_port != db->addr.port)
+		else if (v_port != db->port)
 			changed = true;
 		else if (username && !db->forced_user)
 			changed = true;
@@ -284,8 +287,6 @@ void parse_database(char *name, char *connstr)
 			changed = true;
 		else if (!username && db->forced_user)
 			changed = true;
-		else if (strcmp(db->unix_socket_dir, unix_dir) != 0)
-			changed = true;
 		else if ((db->connect_query && !connect_query)
 			 || (!db->connect_query && connect_query)
 			 || (connect_query && strcmp(connect_query, db->connect_query) != 0))
@@ -297,39 +298,49 @@ void parse_database(char *name, char *connstr)
 
 	/* if pool_size < 0 it will be set later */
 	db->pool_size = pool_size;
-	db->addr.port = v_port;
-	db->addr.ip_addr.s_addr = v_addr;
-	db->addr.is_unix = host ? 0 : 1;
-	safe_strcpy(db->unix_socket_dir, unix_dir, sizeof(db->unix_socket_dir));
+	db->res_pool_size = res_pool_size;
 
-	if (host)
-		log_debug("%s: host=%s/%s", name, host, inet_ntoa(db->addr.ip_addr));
+	if (db->host)
+		free(db->host);
+	db->host = host;
+	db->port = v_port;
 
 	/* assign connect_query */
 	set_connect_query(db, connect_query);
 
-	pktbuf_static(&buf, db->startup_params, sizeof(db->startup_params));
+	if (db->startup_params) {
+		msg = db->startup_params;
+		pktbuf_reset(msg);
+	} else {
+		msg = pktbuf_dynamic(128);
+		if (!msg)
+			fatal("cannot allocate startup buf");
+		db->startup_params = msg;
+	}
 
-	pktbuf_put_string(&buf, "database");
-	db->dbname = (char *)db->startup_params + pktbuf_written(&buf);
-	pktbuf_put_string(&buf, dbname);
+	pktbuf_put_string(msg, "database");
+	dbname_ofs = msg->write_pos;
+	pktbuf_put_string(msg, dbname);
 
 	if (client_encoding) {
-		pktbuf_put_string(&buf, "client_encoding");
-		pktbuf_put_string(&buf, client_encoding);
+		pktbuf_put_string(msg, "client_encoding");
+		pktbuf_put_string(msg, client_encoding);
 	}
 
 	if (datestyle) {
-		pktbuf_put_string(&buf, "datestyle");
-		pktbuf_put_string(&buf, datestyle);
+		pktbuf_put_string(msg, "datestyle");
+		pktbuf_put_string(msg, datestyle);
 	}
 
 	if (timezone) {
-		pktbuf_put_string(&buf, "timezone");
-		pktbuf_put_string(&buf, timezone);
+		pktbuf_put_string(msg, "timezone");
+		pktbuf_put_string(msg, timezone);
 	}
 
-	db->startup_params_len = pktbuf_written(&buf);
+	if (appname) {
+		pktbuf_put_string(msg, "application_name");
+		pktbuf_put_string(msg, appname);
+	}
 
 	/* if user is forces, create fake object for it */
 	if (username != NULL) {
@@ -338,6 +349,14 @@ void parse_database(char *name, char *connstr)
 	} else if (db->forced_user)
 		log_warning("losing forced user not supported,"
 			    " keeping old setting");
+
+	/* remember dbname */
+	db->dbname = (char *)msg->buf + dbname_ofs;
+	free(tmp_connstr);
+	return true;
+fail:
+	free(tmp_connstr);
+	return true;
 }
 
 /*
@@ -345,11 +364,12 @@ void parse_database(char *name, char *connstr)
  */
 
 /* find next " in string, skipping escaped ones */
-static char *find_quote(char *p)
+static char *find_quote(char *p, bool start)
 {
 loop:
-	while (*p && *p != '\\' && *p != '"') p++;
-	if (*p == '\\' && p[1]) {
+	while (*p && *p != '"')
+		p++;
+	if (p[0] == '"' && p[1] == '"' && !start) {
 		p += 2;
 		goto loop;
 	}
@@ -362,10 +382,9 @@ static void copy_quoted(char *dst, const char *src, int len)
 {
 	char *end = dst + len - 1;
 	while (*src && dst < end) {
-		if (*src != '\\')
-			*dst++ = *src++;
-		else
+		if (*src == '"')
 			src++;
+		*dst++ = *src++;
 	}
 	*dst = 0;
 }
@@ -421,7 +440,7 @@ bool loader_users_check(void)
 static void disable_users(void)
 {
 	PgUser *user;
-	List *item;
+	struct List *item;
 
 	statlist_for_each(item, &user_list) {
 		user = container_of(item, PgUser, head);
@@ -434,7 +453,7 @@ bool load_auth_file(const char *fn)
 {
 	char *user, *password, *buf, *p;
 
-	buf = load_file(fn);
+	buf = load_file(fn, NULL);
 	if (buf == NULL) {
 		/* reset file info */
 		auth_loaded(NULL);
@@ -456,7 +475,7 @@ bool load_auth_file(const char *fn)
 			break;
 		}
 		user = ++p;
-		p = find_quote(p);
+		p = find_quote(p, false);
 		if (*p != '"') {
 			log_error("broken auth file");
 			break;
@@ -468,13 +487,13 @@ bool load_auth_file(const char *fn)
 		*p++ = 0; /* tag username end */
 		
 		/* get password */
-		p = find_quote(p);
+		p = find_quote(p, true);
 		if (*p != '"') {
 			log_error("broken auth file");
 			break;
 		}
 		password = ++p;
-		p = find_quote(p);
+		p = find_quote(p, false);
 		if (*p != '"') {
 			log_error("broken auth file");
 			break;
@@ -496,203 +515,3 @@ bool load_auth_file(const char *fn)
 	return true;
 }
 
-/*
- * INI file parser
- */
-
-bool cf_set_int(ConfElem *elem, const char *val, PgSocket *console)
-{
-	int *int_p = elem->dst;
-	if (*val < '0' || *val > '9') {
-		admin_error(console, "bad value: %s", val);
-		return false;
-	}
-	*int_p = atoi(val);
-	return true;
-}
-
-const char *cf_get_int(ConfElem *elem)
-{
-	static char numbuf[32];
-	int val;
-
-	val = *(int *)elem->dst;
-	sprintf(numbuf, "%d", val);
-	return numbuf;
-}
-bool cf_set_time(ConfElem *elem, const char *val, PgSocket *console)
-{
-	usec_t *time_p = elem->dst;
-	if (*val < '0' || *val > '9') {
-		admin_error(console, "bad value: %s", val);
-		return false;
-	}
-	*time_p = USEC * (usec_t)atoi(val);
-	return true;
-}
-
-const char *cf_get_time(ConfElem *elem)
-{
-	static char numbuf[32];
-	usec_t val;
-
-	val = *(usec_t *)elem->dst;
-	sprintf(numbuf, "%d", (int)(val / USEC));
-	return numbuf;
-}
-
-bool cf_set_str(ConfElem *elem, const char *val, PgSocket *console)
-{
-	char **str_p = elem->dst;
-	char *tmp;
-
-	/* don't touch if not changed */
-	if (*str_p && strcmp(*str_p, val) == 0)
-		return true;
-
-	/* if dynamically allocated, free it */
-	if (elem->allocated)
-		free(*str_p);
-
-	tmp = strdup(val);
-	if (!tmp)
-		return false;
-
-	*str_p = tmp;
-	elem->allocated = true;
-	return true;
-}
-
-const char * cf_get_str(ConfElem *elem)
-{
-	return *(char **)elem->dst;
-}
-
-bool set_config_param(ConfElem *elem_list,
-		      const char *key, const char *val,
-		      bool reload, PgSocket *console)
-{
-	ConfElem *desc;
-
-	for (desc = elem_list; desc->name; desc++) {
-		if (strcasecmp(key, desc->name) != 0)
-			continue;
-	
-		/* if reload not allowed, skip it */
-		if (reload && !desc->reloadable) {
-			if (console)
-				admin_error(console,
-					"%s cannot be changed online", key);
-			return false;
-		}
-
-		/* got config, parse it */
-		return desc->io.fn_set(desc, val, console);
-	}
-	return false;
-}
-
-static void map_config(ConfSection *sect, char *key, char *val, bool reload)
-{
-	if (sect == NULL)
-		return;
-
-	if (sect->data_fn)
-		sect->data_fn(key, val);
-	else
-		set_config_param(sect->elem_list, key, val, reload, NULL);
-}
-
-const char *conf_to_text(ConfElem *elem)
-{
-	return elem->io.fn_get(elem);
-}
-
-static ConfSection *find_section(ConfSection *sect, const char *name)
-{
-	for (; sect->name; sect++)
-		if (strcasecmp(sect->name, name) == 0)
-			return sect;
-	log_warning("unknown section in config: %s", name);
-	return NULL;
-}
-
-bool iniparser(const char *fn, ConfSection *sect_list, bool reload)
-{
-	char *buf;
-	char *p, *key, *val;
-	int klen, vlen;
-	ConfSection *cur_section = NULL;
-
-	buf = load_file(fn);
-	if (buf == NULL) {
-		if (!reload)
-			exit(1);
-		else
-			return false;
-	}
-
-	p = buf;
-	while (*p) {
-		/* space at the start of line - including empty lines */
-		while (*p && isspace(*p)) p++;
-
-		/* skip comment lines */
-		if (*p == '#' || *p == ';') {
-			while (*p && *p != '\n') p++;
-			continue;
-		}
-		/* got new section */
-		if (*p == '[') {
-			key = ++p;
-			while (*p && *p != ']' && *p != '\n') p++;
-			if (*p != ']') {
-				log_warning("bad section header");
-				cur_section = NULL;
-				continue;
-			}
-			*p++ = 0;
-
-			cur_section = find_section(sect_list, key);
-			continue;
-		}
-
-		/* done? */
-		if (*p == 0) break;
-
-		/* read key val */
-		key = p;
-		while (*p && (isalnum(*p) || *p == '_')) p++;
-		klen = p - key;
-
-		/* expect '=', skip it */
-		while (*p && (*p == ' ' || *p == '\t')) p++;
-		if (*p != '=') {
-			log_error("syntax error in configuration, stopping loading");
-			break;
-		} else
-			p++;
-		while (*p && (*p == ' ' || *p == '\t')) p++;
-
-		/* now read value */
-		val = p;
-		while (*p && (*p != '\n'))
-			p++;
-		vlen = p - val;
-		/* eat space at end */
-		while (vlen > 0 && isspace(val[vlen - 1]))
-			vlen--;
-
-		/* skip junk */
-		while (*p && isspace(*p)) p++;
-
-		/* our buf is r/w, so take it easy */
-		key[klen] = 0;
-		val[vlen] = 0;
-		map_config(cur_section, key, val, reload);
-	}
-
-	free(buf);
-	return true;
-}
-
