@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -27,7 +27,7 @@ STATLIST(user_list);
 STATLIST(database_list);
 STATLIST(pool_list);
 
-Tree user_tree;
+struct AATree user_tree;
 
 /*
  * client and server objects will be pre-allocated
@@ -36,12 +36,12 @@ Tree user_tree;
  */
 STATLIST(login_client_list);
 
-ObjectCache *server_cache;
-ObjectCache *client_cache;
-ObjectCache *db_cache;
-ObjectCache *pool_cache;
-ObjectCache *user_cache;
-ObjectCache *iobuf_cache;
+struct Slab *server_cache;
+struct Slab *client_cache;
+struct Slab *db_cache;
+struct Slab *pool_cache;
+struct Slab *user_cache;
+struct Slab *iobuf_cache;
 
 /*
  * libevent may still report events when event_del()
@@ -51,16 +51,19 @@ ObjectCache *iobuf_cache;
 static STATLIST(justfree_client_list);
 static STATLIST(justfree_server_list);
 
+/* init autodb idle list */
+STATLIST(autodatabase_idle_list);
+
 /* fast way to get number of active clients */
 int get_active_client_count(void)
 {
-	return objcache_active_count(client_cache);
+	return slab_active_count(client_cache);
 }
 
 /* fast way to get number of active servers */
 int get_active_server_count(void)
 {
-	return objcache_active_count(server_cache);
+	return slab_active_count(server_cache);
 }
 
 static void construct_client(void *obj)
@@ -84,7 +87,7 @@ static void construct_server(void *obj)
 }
 
 /* compare string with PgUser->name, for usage with btree */
-static int user_node_cmp(long userptr, Node *node)
+static int user_node_cmp(uintptr_t userptr, struct AANode *node)
 {
 	const char *name = (const char *)userptr;
 	PgUser *user = container_of(node, PgUser, tree_node);
@@ -94,10 +97,10 @@ static int user_node_cmp(long userptr, Node *node)
 /* initialization before config loading */
 void init_objects(void)
 {
-	tree_init(&user_tree, user_node_cmp, NULL);
-	user_cache = objcache_create("user_cache", sizeof(PgUser), 0, NULL);
-	db_cache = objcache_create("db_cache", sizeof(PgDatabase), 0, NULL);
-	pool_cache = objcache_create("pool_cache", sizeof(PgPool), 0, NULL);
+	aatree_init(&user_tree, user_node_cmp, NULL);
+	user_cache = slab_create("user_cache", sizeof(PgUser), 0, NULL, USUAL_ALLOC);
+	db_cache = slab_create("db_cache", sizeof(PgDatabase), 0, NULL, USUAL_ALLOC);
+	pool_cache = slab_create("pool_cache", sizeof(PgPool), 0, NULL, USUAL_ALLOC);
 
 	if (!user_cache || !db_cache || !pool_cache)
 		fatal("cannot create initial caches");
@@ -112,9 +115,9 @@ static void do_iobuf_reset(void *arg)
 /* initialization after config loading */
 void init_caches(void)
 {
-	server_cache = objcache_create("server_cache", sizeof(PgSocket), 0, construct_server);
-	client_cache = objcache_create("client_cache", sizeof(PgSocket), 0, construct_client);
-	iobuf_cache = objcache_create("iobuf_cache", IOBUF_SIZE, 0, do_iobuf_reset);
+	server_cache = slab_create("server_cache", sizeof(PgSocket), 0, construct_server, USUAL_ALLOC);
+	client_cache = slab_create("client_cache", sizeof(PgSocket), 0, construct_client, USUAL_ALLOC);
+	iobuf_cache = slab_create("iobuf_cache", IOBUF_SIZE, 0, do_iobuf_reset, USUAL_ALLOC);
 }
 
 /* state change means moving between lists */
@@ -127,19 +130,19 @@ void change_client_state(PgSocket *client, SocketState newstate)
 	case CL_FREE:
 		break;
 	case CL_JUSTFREE:
-		statlist_remove(&client->head, &justfree_client_list);
+		statlist_remove(&justfree_client_list, &client->head);
 		break;
 	case CL_LOGIN:
-		statlist_remove(&client->head, &login_client_list);
+		statlist_remove(&login_client_list, &client->head);
 		break;
 	case CL_WAITING:
-		statlist_remove(&client->head, &pool->waiting_client_list);
+		statlist_remove(&pool->waiting_client_list, &client->head);
 		break;
 	case CL_ACTIVE:
-		statlist_remove(&client->head, &pool->active_client_list);
+		statlist_remove(&pool->active_client_list, &client->head);
 		break;
 	case CL_CANCEL:
-		statlist_remove(&client->head, &pool->cancel_req_list);
+		statlist_remove(&pool->cancel_req_list, &client->head);
 		break;
 	default:
 		fatal("bad cur client state: %d", client->state);
@@ -150,22 +153,23 @@ void change_client_state(PgSocket *client, SocketState newstate)
 	/* put to new location */
 	switch (client->state) {
 	case CL_FREE:
-		obj_free(client_cache, client);
+		varcache_clean(&client->vars);
+		slab_free(client_cache, client);
 		break;
 	case CL_JUSTFREE:
-		statlist_append(&client->head, &justfree_client_list);
+		statlist_append(&justfree_client_list, &client->head);
 		break;
 	case CL_LOGIN:
-		statlist_append(&client->head, &login_client_list);
+		statlist_append(&login_client_list, &client->head);
 		break;
 	case CL_WAITING:
-		statlist_append(&client->head, &pool->waiting_client_list);
+		statlist_append(&pool->waiting_client_list, &client->head);
 		break;
 	case CL_ACTIVE:
-		statlist_append(&client->head, &pool->active_client_list);
+		statlist_append(&pool->active_client_list, &client->head);
 		break;
 	case CL_CANCEL:
-		statlist_append(&client->head, &pool->cancel_req_list);
+		statlist_append(&pool->cancel_req_list, &client->head);
 		break;
 	default:
 		fatal("bad new client state: %d", client->state);
@@ -182,22 +186,22 @@ void change_server_state(PgSocket *server, SocketState newstate)
 	case SV_FREE:
 		break;
 	case SV_JUSTFREE:
-		statlist_remove(&server->head, &justfree_server_list);
+		statlist_remove(&justfree_server_list, &server->head);
 		break;
 	case SV_LOGIN:
-		statlist_remove(&server->head, &pool->new_server_list);
+		statlist_remove(&pool->new_server_list, &server->head);
 		break;
 	case SV_USED:
-		statlist_remove(&server->head, &pool->used_server_list);
+		statlist_remove(&pool->used_server_list, &server->head);
 		break;
 	case SV_TESTED:
-		statlist_remove(&server->head, &pool->tested_server_list);
+		statlist_remove(&pool->tested_server_list, &server->head);
 		break;
 	case SV_IDLE:
-		statlist_remove(&server->head, &pool->idle_server_list);
+		statlist_remove(&pool->idle_server_list, &server->head);
 		break;
 	case SV_ACTIVE:
-		statlist_remove(&server->head, &pool->active_server_list);
+		statlist_remove(&pool->active_server_list, &server->head);
 		break;
 	default:
 		fatal("change_server_state: bad old server state: %d", server->state);
@@ -208,31 +212,32 @@ void change_server_state(PgSocket *server, SocketState newstate)
 	/* put to new location */
 	switch (server->state) {
 	case SV_FREE:
-		obj_free(server_cache, server);
+		varcache_clean(&server->vars);
+		slab_free(server_cache, server);
 		break;
 	case SV_JUSTFREE:
-		statlist_append(&server->head, &justfree_server_list);
+		statlist_append(&justfree_server_list, &server->head);
 		break;
 	case SV_LOGIN:
-		statlist_append(&server->head, &pool->new_server_list);
+		statlist_append(&pool->new_server_list, &server->head);
 		break;
 	case SV_USED:
 		/* use LIFO */
-		statlist_prepend(&server->head, &pool->used_server_list);
+		statlist_prepend(&pool->used_server_list, &server->head);
 		break;
 	case SV_TESTED:
-		statlist_append(&server->head, &pool->tested_server_list);
+		statlist_append(&pool->tested_server_list, &server->head);
 		break;
 	case SV_IDLE:
 		if (server->close_needed || cf_server_round_robin)
 			/* try to avoid immediate usage then */
-			statlist_append(&server->head, &pool->idle_server_list);
+			statlist_append(&pool->idle_server_list, &server->head);
 		else
 			/* otherwise use LIFO */
-			statlist_prepend(&server->head, &pool->idle_server_list);
+			statlist_prepend(&pool->idle_server_list, &server->head);
 		break;
 	case SV_ACTIVE:
-		statlist_append(&server->head, &pool->active_server_list);
+		statlist_append(&pool->active_server_list, &server->head);
 		break;
 	default:
 		fatal("bad server state");
@@ -240,7 +245,7 @@ void change_server_state(PgSocket *server, SocketState newstate)
 }
 
 /* compare pool names, for use with put_in_order */
-static int cmp_pool(List *i1, List *i2)
+static int cmp_pool(struct List *i1, struct List *i2)
 {
 	PgPool *p1 = container_of(i1, PgPool, head);
 	PgPool *p2 = container_of(i2, PgPool, head);
@@ -252,7 +257,7 @@ static int cmp_pool(List *i1, List *i2)
 }
 
 /* compare user names, for use with put_in_order */
-static int cmp_user(List *i1, List *i2)
+static int cmp_user(struct List *i1, struct List *i2)
 {
 	PgUser *u1 = container_of(i1, PgUser, head);
 	PgUser *u2 = container_of(i2, PgUser, head);
@@ -260,7 +265,7 @@ static int cmp_user(List *i1, List *i2)
 }
 
 /* compare db names, for use with put_in_order */
-static int cmp_database(List *i1, List *i2)
+static int cmp_database(struct List *i1, struct List *i2)
 {
 	PgDatabase *db1 = container_of(i1, PgDatabase, head);
 	PgDatabase *db2 = container_of(i2, PgDatabase, head);
@@ -268,21 +273,22 @@ static int cmp_database(List *i1, List *i2)
 }
 
 /* put elem into list in correct pos */
-static void put_in_order(List *newitem, StatList *list, int (*cmpfn)(List *, List *))
+static void put_in_order(struct List *newitem, struct StatList *list,
+			 int (*cmpfn)(struct List *, struct List *))
 {
 	int res;
-	List *item;
+	struct List *item;
 
 	statlist_for_each(item, list) {
 		res = cmpfn(item, newitem);
 		if (res == 0)
 			fatal("put_in_order: found existing elem");
 		else if (res > 0) {
-			statlist_put_before(newitem, list, item);
+			statlist_put_before(list, newitem, item);
 			return;
 		}
 	}
-	statlist_append(newitem, list);
+	statlist_append(list, newitem);
 }
 
 /* create new object if new, then return it */
@@ -292,7 +298,7 @@ PgDatabase *add_database(const char *name)
 
 	/* create new object if needed */
 	if (db == NULL) {
-		db = obj_alloc(db_cache);
+		db = slab_alloc(db_cache);
 		if (!db)
 			return NULL;
 
@@ -304,13 +310,44 @@ PgDatabase *add_database(const char *name)
 	return db;
 }
 
+/* register new auto database */
+PgDatabase *register_auto_database(const char *name)
+{
+	PgDatabase *db;
+	int len;
+	char *cs;
+	
+	if (!cf_autodb_connstr)
+		return NULL;
+
+	len = strlen(cf_autodb_connstr);
+	cs = malloc(len + 1);
+	if (!cs)
+		return NULL;
+	memcpy(cs, cf_autodb_connstr, len + 1);
+	parse_database(NULL, (char*)name, cs);
+	free(cs);
+
+	db = find_database(name);
+	if (db) {
+		db->db_auto = 1;
+		/* do not forget to check pool_size like in config_postprocess */
+		if (db->pool_size < 0)
+			db->pool_size = cf_default_pool_size;
+		if (db->res_pool_size < 0)
+			db->res_pool_size = cf_res_pool_size;
+	}
+
+	return db;
+}
+
 /* add or update client users */
 PgUser *add_user(const char *name, const char *passwd)
 {
 	PgUser *user = find_user(name);
 
 	if (user == NULL) {
-		user = obj_alloc(user_cache);
+		user = slab_alloc(user_cache);
 		if (!user)
 			return NULL;
 
@@ -319,7 +356,7 @@ PgUser *add_user(const char *name, const char *passwd)
 		safe_strcpy(user->name, name, sizeof(user->name));
 		put_in_order(&user->head, &user_list, cmp_user);
 
-		tree_insert(&user_tree, (long)user->name, &user->tree_node);
+		aatree_insert(&user_tree, (uintptr_t)user->name, &user->tree_node);
 	}
 	safe_strcpy(user->passwd, passwd, sizeof(user->passwd));
 	return user;
@@ -330,7 +367,7 @@ PgUser *force_user(PgDatabase *db, const char *name, const char *passwd)
 {
 	PgUser *user = db->forced_user;
 	if (!user) {
-		user = obj_alloc(user_cache);
+		user = slab_alloc(user_cache);
 		if (!user)
 			return NULL;
 		list_init(&user->head);
@@ -345,13 +382,23 @@ PgUser *force_user(PgDatabase *db, const char *name, const char *passwd)
 /* find an existing database */
 PgDatabase *find_database(const char *name)
 {
-	List *item;
+	struct List *item, *tmp;
 	PgDatabase *db;
 	statlist_for_each(item, &database_list) {
 		db = container_of(item, PgDatabase, head);
 		if (strcmp(db->name, name) == 0)
 			return db;
 	}
+	/* also trying to find in idle autodatabases list */
+	statlist_for_each_safe(item, &autodatabase_idle_list, tmp) {
+		db = container_of(item, PgDatabase, head);
+		if (strcmp(db->name, name) == 0) {
+			db->inactive_time = 0;
+			statlist_remove(&autodatabase_idle_list, &db->head);
+			put_in_order(&db->head, &database_list, cmp_database);
+			return db;
+		}
+	}
 	return NULL;
 }
 
@@ -359,9 +406,9 @@ PgDatabase *find_database(const char *name)
 PgUser *find_user(const char *name)
 {
 	PgUser *user = NULL;
-	Node *node;
+	struct AANode *node;
 
-	node = tree_search(&user_tree, (long)name);
+	node = aatree_search(&user_tree, (uintptr_t)name);
 	user = node ? container_of(node, PgUser, tree_node) : NULL;
 	return user;
 }
@@ -371,7 +418,7 @@ static PgPool *new_pool(PgDatabase *db, PgUser *user)
 {
 	PgPool *pool;
 
-	pool = obj_alloc(pool_cache);
+	pool = slab_alloc(pool_cache);
 	if (!pool)
 		return NULL;
 
@@ -390,7 +437,7 @@ static PgPool *new_pool(PgDatabase *db, PgUser *user)
 	statlist_init(&pool->new_server_list, "new_server_list");
 	statlist_init(&pool->cancel_req_list, "cancel_req_list");
 
-	list_append(&pool->map_head, &user->pool_list);
+	list_append(&user->pool_list, &pool->map_head);
 
 	/* keep pools in db/user order to make stats faster */
 	put_in_order(&pool->head, &pool_list, cmp_pool);
@@ -401,7 +448,7 @@ static PgPool *new_pool(PgDatabase *db, PgUser *user)
 /* find pool object, create if needed */
 PgPool *get_pool(PgDatabase *db, PgUser *user)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 
 	if (!db || !user)
@@ -437,6 +484,39 @@ void activate_client(PgSocket *client)
 	sbuf_continue(&client->sbuf);
 }
 
+/*
+ * Don't let clients queue at all, if there is no working server connection.
+ *
+ * It must still allow following cases:
+ * - empty pool on startup
+ * - idle pool where all servers are removed
+ *
+ * Current assumptions:
+ * - old server connections will be dropped by query_timeout
+ * - new server connections fail due to server_connect_timeout, or other failure
+ *
+ * So here we drop client if all server connections have been dropped
+ * and new one's fail.
+ */
+bool check_fast_fail(PgSocket *client)
+{
+	int cnt;
+	PgPool *pool = client->pool;
+
+	/* reject if no servers and last connect failed */
+	if (!pool->last_connect_failed)
+		return true;
+	cnt = pool_server_count(pool) - statlist_count(&pool->new_server_list);
+	if (cnt)
+		return true;
+	disconnect_client(client, true, "no working server connection");
+
+	/* usual relaunch wont work, as there are no waiting clients */
+	launch_new_connection(pool);
+
+	return false;
+}
+
 /* link if found, otherwise put into wait queue */
 bool find_server(PgSocket *client)
 {
@@ -456,10 +536,19 @@ bool find_server(PgSocket *client)
 	} else {
 		while (1) {
 			server = first_socket(&pool->idle_server_list);
-			if (!server || server->ready)
+			if (!server)
+				break;
+			else if (server->close_needed)
+				disconnect_server(server, true, "obsolete connection");
+			else if (!server->ready)
+				disconnect_server(server, true, "idle server got dirty");
+			else
 				break;
-			disconnect_server(server, true, "idle server got dirty");
 		}
+
+		if (!server && !check_fast_fail(client))
+			return false;
+
 	}
 	Assert(!server || server->state == SV_IDLE);
 
@@ -577,12 +666,13 @@ bool release_server(PgSocket *server)
 		pool->last_connect_failed = 0;
 		break;
 	default:
-		fatal("bad server state in release_server");
+		fatal("bad server state in release_server (%d)", server->state);
 	}
 
 	/* enforce lifetime immidiately on release */
 	if (server->state != SV_LOGIN && life_over(server)) {
 		disconnect_server(server, true, "server_lifetime");
+		pool->last_lifetime_disconnect = get_cached_time();
 		return false;
 	}
 
@@ -606,16 +696,23 @@ bool release_server(PgSocket *server)
 }
 
 /* drop server connection */
-void disconnect_server(PgSocket *server, bool notify, const char *reason)
+void disconnect_server(PgSocket *server, bool notify, const char *reason, ...)
 {
 	PgPool *pool = server->pool;
-	PgSocket *client = server->link;
+	PgSocket *client;
 	static const uint8_t pkt_term[] = {'X', 0,0,0,4};
 	int send_term = 1;
 	usec_t now = get_cached_time();
+	char buf[128];
+	va_list ap;
+
+	va_start(ap, reason);
+	vsnprintf(buf, sizeof(buf), reason, ap);
+	va_end(ap);
+	reason = buf;
 
 	if (cf_log_disconnections)
-		slog_info(server, "closing because: %s (age=%llu)", reason,
+		slog_info(server, "closing because: %s (age=%" PRIu64 ")", reason,
 			  (now - server->connect_time) / USEC);
 
 	switch (server->state) {
@@ -624,7 +721,7 @@ void disconnect_server(PgSocket *server, bool notify, const char *reason)
 		if (client) {
 			client->link = NULL;
 			server->link = NULL;
-			disconnect_client(client, true, reason);
+			disconnect_client(client, true, "%s", reason);
 		}
 		break;
 	case SV_TESTED:
@@ -642,7 +739,7 @@ void disconnect_server(PgSocket *server, bool notify, const char *reason)
 			send_term = 0;
 		break;
 	default:
-		fatal("disconnect_server: bad server state");
+		fatal("disconnect_server: bad server state (%d)", server->state);
 	}
 
 	Assert(server->link == NULL);
@@ -654,18 +751,30 @@ void disconnect_server(PgSocket *server, bool notify, const char *reason)
 			notify = false;
 	}
 
+	if (server->dns_token) {
+		adns_cancel(adns, server->dns_token);
+		server->dns_token = NULL;
+	}
+
 	change_server_state(server, SV_JUSTFREE);
 	if (!sbuf_close(&server->sbuf))
 		log_noise("sbuf_close failed, retry later");
 }
 
 /* drop client connection */
-void disconnect_client(PgSocket *client, bool notify, const char *reason)
+void disconnect_client(PgSocket *client, bool notify, const char *reason, ...)
 {
+	char buf[128];
+	va_list ap;
 	usec_t now = get_cached_time();
 
+	va_start(ap, reason);
+	vsnprintf(buf, sizeof(buf), reason, ap);
+	va_end(ap);
+	reason = buf;
+
 	if (cf_log_disconnections)
-		slog_info(client, "closing because: %s (age=%llu)", reason,
+		slog_info(client, "closing because: %s (age=%" PRIu64 ")", reason,
 			  (now - client->connect_time) / USEC);
 
 	switch (client->state) {
@@ -704,13 +813,118 @@ void disconnect_client(PgSocket *client, bool notify, const char *reason)
 		log_noise("sbuf_close failed, retry later");
 }
 
+/*
+ * Connection creation utilities
+ */
+
+static void connect_server(struct PgSocket *server, const struct sockaddr *sa, int salen)
+{
+	bool res;
+
+	/* fill remote_addr */
+	memset(&server->remote_addr, 0, sizeof(server->remote_addr));
+	if (sa->sa_family == AF_UNIX) {
+		pga_set(&server->remote_addr, AF_UNIX, server->pool->db->port);
+	} else {
+		pga_copy(&server->remote_addr, sa);
+	}
+
+	if (cf_log_connections)
+		slog_info(server, "new connection to server");
+
+	/* start connecting */
+	res = sbuf_connect(&server->sbuf, sa, salen,
+			   cf_server_connect_timeout / USEC);
+	if (!res)
+		log_noise("failed to launch new connection");
+}
+
+static void dns_callback(void *arg, const struct sockaddr *sa, int salen)
+{
+	struct PgSocket *server = arg;
+	struct PgDatabase *db = server->pool->db;
+	struct sockaddr_in sa_in;
+
+	server->dns_token = NULL;
+
+	if (!sa) {
+		disconnect_server(server, true, "server dns lookup failed");
+		return;
+	} else if (sa->sa_family == AF_INET) {
+		char buf[64];
+		memcpy(&sa_in, sa, sizeof(sa_in));
+		sa_in.sin_port = htons(db->port);
+		sa = (struct sockaddr *)&sa_in;
+		salen = sizeof(sa_in);
+		slog_debug(server, "dns_callback: inet4: %s",
+			   sa2str(sa, buf, sizeof(buf)));
+	} else {
+		disconnect_server(server, true, "unknown dns type");
+		return;
+	}
+
+	connect_server(server, sa, salen);
+}
+
+static void dns_connect(struct PgSocket *server)
+{
+	struct sockaddr_un sa_un;
+	struct sockaddr_in sa_in;
+	struct sockaddr_in6 sa_in6;
+	struct sockaddr *sa;
+	struct PgDatabase *db = server->pool->db;
+	const char *host = db->host;
+	const char *unix_dir;
+	int sa_len;
+
+	if (!host || host[0] == '/') {
+		slog_noise(server, "unix socket: %s", sa_un.sun_path);
+		memset(&sa_un, 0, sizeof(sa_un));
+		sa_un.sun_family = AF_UNIX;
+		unix_dir = host ? host : cf_unix_socket_dir;
+		if (!unix_dir || !*unix_dir) {
+			log_error("Unix socket dir not configured: %s", db->name);
+			disconnect_server(server, false, "cannot connect");
+			return;
+		}
+		snprintf(sa_un.sun_path, sizeof(sa_un.sun_path),
+			 "%s/.s.PGSQL.%d", unix_dir, db->port);
+		sa = (struct sockaddr *)&sa_un;
+		sa_len = sizeof(sa_un);
+	} else if (strchr(host, ':')) {  // assume IPv6 address on any : in addr
+		slog_noise(server, "inet6 socket: %s", db->host);
+		memset(&sa_in6, 0, sizeof(sa_in6));
+		sa_in6.sin6_family = AF_INET6;
+		inet_pton(AF_INET6, db->host, (void *) sa_in6.sin6_addr.s6_addr);
+		sa_in6.sin6_port = htons(db->port);
+		sa = (struct sockaddr *)&sa_in6;
+		sa_len = sizeof(sa_in6);
+	} else if (host[0] >= '0' && host[0] <= '9') { // else try IPv4
+		slog_noise(server, "inet socket: %s", db->host);
+		memset(&sa_in, 0, sizeof(sa_in));
+		sa_in.sin_family = AF_INET;
+		sa_in.sin_addr.s_addr = inet_addr(db->host);
+		sa_in.sin_port = htons(db->port);
+		sa = (struct sockaddr *)&sa_in;
+		sa_len = sizeof(sa_in);
+	} else {
+		struct DNSToken *tk;
+		slog_noise(server, "dns socket: %s", db->host);
+		/* launch dns lookup */
+		tk = adns_resolve(adns, db->host, dns_callback, server);
+		if (tk)
+			server->dns_token = tk;
+		return;
+	}
+
+	connect_server(server, sa, sa_len);
+}
+
 /* the pool needs new connection, if possible */
 void launch_new_connection(PgPool *pool)
 {
 	PgSocket *server;
 	int total;
-	const char *unix_dir = cf_unix_socket_dir;
-	bool res;
 
 	/* allow only small number of connection attempts at a time */
 	if (!statlist_empty(&pool->new_server_list)) {
@@ -730,13 +944,25 @@ void launch_new_connection(PgPool *pool)
 	/* is it allowed to add servers? */
 	total = pool_server_count(pool);
 	if (total >= pool->db->pool_size && pool->welcome_msg_ready) {
+		/* should we use reserve pool? */
+		if (cf_res_pool_timeout && pool->db->res_pool_size) {
+			usec_t now = get_cached_time();
+			PgSocket *c = first_socket(&pool->waiting_client_list);
+			if (c && (now - c->request_time) >= cf_res_pool_timeout) {
+				if (total < pool->db->pool_size + pool->db->res_pool_size) {
+					slog_warning(c, "Taking connection from reserve_pool");
+					goto allow_new;
+				}
+			}
+		}
 		log_debug("launch_new_connection: pool full (%d >= %d)",
 				total, pool->db->pool_size);
 		return;
 	}
 
+allow_new:
 	/* get free conn object */
-	server = obj_alloc(server_cache);
+	server = slab_alloc(server_cache);
 	if (!server) {
 		log_debug("launch_new_connection: no memory");
 		return;
@@ -745,35 +971,21 @@ void launch_new_connection(PgPool *pool)
 	/* initialize it */
 	server->pool = pool;
 	server->auth_user = server->pool->user;
-	server->remote_addr = server->pool->db->addr;
 	server->connect_time = get_cached_time();
 	pool->last_connect_time = get_cached_time();
 	change_server_state(server, SV_LOGIN);
 
-	if (cf_log_connections)
-		slog_info(server, "new connection to server");
-
-	/* override socket location if requested */
-	if (server->pool->db->unix_socket_dir[0])
-		unix_dir = server->pool->db->unix_socket_dir;
-
-	/* start connecting */
-	res = sbuf_connect(&server->sbuf, &server->remote_addr, unix_dir,
-			   cf_server_connect_timeout / USEC);
-	if (!res)
-		log_noise("failed to launch new connection");
+	dns_connect(server);
 }
 
 /* new client connection attempt */
-PgSocket * accept_client(int sock,
-			 const struct sockaddr_in *addr,
-			 bool is_unix)
+PgSocket *accept_client(int sock, bool is_unix)
 {
 	bool res;
 	PgSocket *client;
 
 	/* get free PgSocket */
-	client = obj_alloc(client_cache);
+	client = slab_alloc(client_cache);
 	if (!client) {
 		log_warning("cannot allocate client struct");
 		safe_close(sock);
@@ -783,6 +995,7 @@ PgSocket * accept_client(int sock,
 	client->connect_time = client->request_time = get_cached_time();
 	client->query_start = 0;
 
+	/* FIXME: take local and remote address from pool_accept() */
 	fill_remote_addr(client, sock, is_unix);
 	fill_local_addr(client, sock, is_unix);
 
@@ -811,7 +1024,8 @@ bool finish_client_login(PgSocket *client)
 		fatal("bad client state");
 	}
 
-	if (!welcome_client(client)) {
+	/* check if we know server signature */
+	if (!client->pool->welcome_msg_ready) {
 		log_debug("finish_client_login: no welcome message, pause");
 		client->wait_for_welcome = 1;
 		pause_client(client);
@@ -821,6 +1035,10 @@ bool finish_client_login(PgSocket *client)
 	}
 	client->wait_for_welcome = 0;
 
+	/* send the message */
+	if (!welcome_client(client))
+		return false;
+
 	slog_debug(client, "logged in");
 
 	return true;
@@ -829,8 +1047,8 @@ bool finish_client_login(PgSocket *client)
 /* client->cancel_key has requested client key */
 void accept_cancel_request(PgSocket *req)
 {
-	List *pitem, *citem;
-	PgPool *pool;
+	struct List *pitem, *citem;
+	PgPool *pool = NULL;
 	PgSocket *server = NULL, *client, *main_client = NULL;
 
 	Assert(req->state == CL_LOGIN);
@@ -870,8 +1088,7 @@ found:
 		/* notify readiness */
 		SEND_ReadyForQuery(res, main_client);
 		if (!res)
-			disconnect_client(main_client, true,
-					  "ReadyForQuery for main_client failed");
+			disconnect_client(main_client, true, "ReadyForQuery for main_client failed");
 		return;
 	}
 
@@ -913,7 +1130,7 @@ bool use_client_socket(int fd, PgAddr *addr,
 	PgSocket *client;
 	PktBuf tmp;
 
-	client = accept_client(fd, NULL, addr->is_unix);
+	client = accept_client(fd, pga_is_unix(addr));
 	if (client == NULL)
 		return false;
 	client->suspended = 1;
@@ -951,6 +1168,13 @@ bool use_server_socket(int fd, PgAddr *addr,
 	PgSocket *server;
 	PktBuf tmp;
 	bool res;
+	
+	/* if the database not found, it's an auto database -> registering... */
+	if (!db) {
+		db = register_auto_database(dbname);
+		if (!db)
+			return true;
+	}
 
 	if (db->forced_user)
 		user = db->forced_user;
@@ -961,11 +1185,11 @@ bool use_server_socket(int fd, PgAddr *addr,
 	if (!pool)
 		return false;
 
-	server = obj_alloc(server_cache);
+	server = slab_alloc(server_cache);
 	if (!server)
 		return false;
 
-	res = sbuf_accept(&server->sbuf, fd, addr->is_unix);
+	res = sbuf_accept(&server->sbuf, fd, pga_is_unix(addr));
 	if (!res)
 		return false;
 
@@ -975,8 +1199,8 @@ bool use_server_socket(int fd, PgAddr *addr,
 	server->connect_time = server->request_time = get_cached_time();
 	server->query_start = 0;
 
-	fill_remote_addr(server, fd, addr->is_unix);
-	fill_local_addr(server, fd, addr->is_unix);
+	fill_remote_addr(server, fd, pga_is_unix(addr));
+	fill_local_addr(server, fd, pga_is_unix(addr));
 
 	if (linkfd) {
 		server->ready = 0;
@@ -1004,7 +1228,7 @@ bool use_server_socket(int fd, PgAddr *addr,
 
 void for_each_server(PgPool *pool, void (*func)(PgSocket *sk))
 {
-	List *item;
+	struct List *item;
 
 	statlist_for_each(item, &pool->idle_server_list)
 		func(container_of(item, PgSocket, head));
@@ -1029,7 +1253,7 @@ static void tag_dirty(PgSocket *sk)
 
 void tag_database_dirty(PgDatabase *db)
 {
-	List *item;
+	struct List *item;
 	PgPool *pool;
 
 	statlist_for_each(item, &pool_list) {
@@ -1042,7 +1266,7 @@ void tag_database_dirty(PgDatabase *db)
 /* move objects from justfree_* to free_* lists */
 void reuse_just_freed_objects(void)
 {
-	List *tmp, *item;
+	struct List *tmp, *item;
 	PgSocket *sk;
 	bool close_works = true;
 
