@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -23,48 +23,62 @@
 #include "bouncer.h"
 
 /*
- * parse protocol header from MBuf
+ * parse protocol header from struct MBuf
  */
 
 /* parses pkt header from buffer, returns false if failed */
-bool get_header(MBuf *data, PktHdr *pkt)
+bool get_header(struct MBuf *data, PktHdr *pkt)
 {
 	unsigned type;
-	unsigned len;
-	unsigned code;
+	uint32_t len;
 	unsigned got;
 	unsigned avail;
-	MBuf hdr;
+	uint16_t len16;
+	uint8_t type8;
+	uint32_t code;
+	struct MBuf hdr;
+	const uint8_t *ptr;
 
 	mbuf_copy(data, &hdr);
 
-	if (mbuf_avail(&hdr) < NEW_HEADER_LEN) {
+	if (mbuf_avail_for_read(&hdr) < NEW_HEADER_LEN) {
 		log_noise("get_header: less then 5 bytes available");
 		return false;
 	}
-	type = mbuf_get_char(&hdr);
+	if (!mbuf_get_byte(&hdr, &type8))
+		return false;
+	type = type8;
 	if (type != 0) {
 		/* wire length does not include type byte */
-		len = mbuf_get_uint32(&hdr) + 1;
+		if (!mbuf_get_uint32be(&hdr, &len))
+			return false;
+		len++;
 		got = NEW_HEADER_LEN;
 	} else {
-		if (mbuf_get_char(&hdr) != 0) {
+		if (!mbuf_get_byte(&hdr, &type8))
+			return false;
+		if (type8 != 0) {
 			log_noise("get_header: unknown special pkt");
 			return false;
 		}
 		/* dont tolerate partial pkt */
-		if (mbuf_avail(&hdr) < OLD_HEADER_LEN - 2) {
+		if (mbuf_avail_for_read(&hdr) < OLD_HEADER_LEN - 2) {
 			log_noise("get_header: less than 8 bytes for special pkt");
 			return false;
 		}
-		len = mbuf_get_uint16(&hdr);
-		code = mbuf_get_uint32(&hdr);
+		if (!mbuf_get_uint16be(&hdr, &len16))
+			return false;
+		len = len16;
+		if (!mbuf_get_uint32be(&hdr, &code))
+			return false;
 		if (code == PKT_CANCEL)
 			type = PKT_CANCEL;
 		else if (code == PKT_SSLREQ)
 			type = PKT_SSLREQ;
 		else if ((code >> 16) == 3 && (code & 0xFFFF) < 2)
 			type = PKT_STARTUP;
+		else if (code == PKT_STARTUP_V2)
+			type = PKT_STARTUP_V2;
 		else {
 			log_noise("get_header: unknown special pkt: len=%u code=%u", len, code);
 			return false;
@@ -81,16 +95,15 @@ bool get_header(MBuf *data, PktHdr *pkt)
 	pkt->len = len;
 
 	/* fill pkt with only data for this packet */
-	if (len > mbuf_avail(data))
-		avail = mbuf_avail(data);
+	if (len > mbuf_avail_for_read(data))
+		avail = mbuf_avail_for_read(data);
 	else
 		avail = len;
-	mbuf_slice(data, avail, &pkt->data);
+	if (!mbuf_slice(data, avail, &pkt->data))
+		return false;
 
 	/* tag header as read */
-	mbuf_get_bytes(&pkt->data, got);
-
-	return true;
+	return mbuf_get_bytes(&pkt->data, got, &ptr);
 }
 
 
@@ -104,7 +117,7 @@ bool send_pooler_error(PgSocket *client, bool send_ready, const char *msg)
 	PktBuf buf;
 
 	if (cf_log_pooler_errors)
-		slog_info(client, "Pooler Error: %s", msg);
+		slog_warning(client, "Pooler Error: %s", msg);
 
 	pktbuf_static(&buf, tmpbuf, sizeof(tmpbuf));
 	pktbuf_write_generic(&buf, 'E', "cscscsc",
@@ -120,13 +133,13 @@ bool send_pooler_error(PgSocket *client, bool send_ready, const char *msg)
 void parse_server_error(PktHdr *pkt, const char **level_p, const char **msg_p)
 {
 	const char *level = NULL, *msg = NULL, *val;
-	int type;
-	while (mbuf_avail(&pkt->data)) {
-		type = mbuf_get_char(&pkt->data);
+	uint8_t type;
+	while (mbuf_avail_for_read(&pkt->data)) {
+		if (!mbuf_get_byte(&pkt->data, &type))
+			break;
 		if (type == 0)
 			break;
-		val = mbuf_get_string(&pkt->data);
-		if (!val)
+		if (!mbuf_get_string(&pkt->data, &val))
 			break;
 		if (type == 'S')
 			level = val;
@@ -155,25 +168,29 @@ void log_server_error(const char *note, PktHdr *pkt)
  */
 
 /* add another server parameter packet to cache */
-void add_welcome_parameter(PgPool *pool, const char *key, const char *val)
+bool add_welcome_parameter(PgPool *pool, const char *key, const char *val)
 {
-	PktBuf msg;
+	PktBuf *msg = pool->welcome_msg;
 
 	if (pool->welcome_msg_ready)
-		return;
+		return true;
 
-	pktbuf_static(&msg, pool->welcome_msg + pool->welcome_msg_len,
-		      sizeof(pool->welcome_msg) - pool->welcome_msg_len);
+	if (!msg) {
+		msg = pktbuf_dynamic(128);
+		if (!msg)
+			return false;
+		pool->welcome_msg = msg;
+	}
 
 	/* first packet must be AuthOk */
-	if (pool->welcome_msg_len == 0)
-		pktbuf_write_AuthenticationOk(&msg);
+	if (msg->write_pos == 0)
+		pktbuf_write_AuthenticationOk(msg);
 
 	/* if not stored in ->orig_vars, write full packet */
-	if (!varcache_set(&pool->orig_vars, key, val)) {
-		pktbuf_write_ParameterStatus(&msg, key, val);
-		pool->welcome_msg_len += pktbuf_written(&msg);
-	}
+	if (!varcache_set(&pool->orig_vars, key, val))
+		pktbuf_write_ParameterStatus(msg, key, val);
+
+	return !msg->failed;
 }
 
 /* all parameters processed */
@@ -188,30 +205,37 @@ void finish_welcome_msg(PgSocket *server)
 bool welcome_client(PgSocket *client)
 {
 	int res;
-	uint8_t buf[1024];
-	PktBuf msg;
 	PgPool *pool = client->pool;
+	const PktBuf *pmsg = pool->welcome_msg;
+	PktBuf *msg;
 
 	slog_noise(client, "P: welcome_client");
-	if (!pool->welcome_msg_ready)
-		return false;
 
-	pktbuf_static(&msg, buf, sizeof(buf));
-	pktbuf_put_bytes(&msg, pool->welcome_msg, pool->welcome_msg_len);
+	/* copy prepared stuff around */
+	msg = pktbuf_temp();
+	pktbuf_put_bytes(msg, pmsg->buf, pmsg->write_pos);
 
+	/* fill vars */
 	varcache_fill_unset(&pool->orig_vars, client);
-	varcache_add_params(&msg, &client->vars);
+	varcache_add_params(msg, &client->vars);
 
 	/* give each client its own cancel key */
 	get_random_bytes(client->cancel_key, 8);
-	pktbuf_write_BackendKeyData(&msg, client->cancel_key);
-	pktbuf_write_ReadyForQuery(&msg);
+	pktbuf_write_BackendKeyData(msg, client->cancel_key);
+
+	/* finish */
+	pktbuf_write_ReadyForQuery(msg);
+	if (msg->failed) {
+		disconnect_client(client, true, "failed to prepare welcome message");
+		return false;
+	}
 
 	/* send all together */
-	res = pktbuf_send_immidiate(&msg, client);
-	if (!res)
-		slog_warning(client, "unhandled failure to send welcome_msg");
-
+	res = pktbuf_send_immidiate(msg, client);
+	if (!res) {
+		disconnect_client(client, true, "failed to send welcome message");
+		return false;
+	}
 	return true;
 }
 
@@ -243,6 +267,10 @@ static bool login_crypt_psw(PgSocket *server, const uint8_t *salt)
 	memcpy(saltbuf, salt, 2);
 	saltbuf[2] = 0;
 	enc = crypt(user->passwd, saltbuf);
+	if (!enc) {
+		slog_warning(server, "crypt failed");
+		return false;
+	}
 	return send_password(server, enc);
 }
 
@@ -265,15 +293,16 @@ static bool login_md5_psw(PgSocket *server, const uint8_t *salt)
 /* answer server authentication request */
 bool answer_authreq(PgSocket *server, PktHdr *pkt)
 {
-	unsigned cmd;
+	uint32_t cmd;
 	const uint8_t *salt;
 	bool res = false;
 
 	/* authreq body must contain 32bit cmd */
-	if (mbuf_avail(&pkt->data) < 4)
+	if (mbuf_avail_for_read(&pkt->data) < 4)
 		return false;
 
-	cmd = mbuf_get_uint32(&pkt->data);
+	if (!mbuf_get_uint32be(&pkt->data, &cmd))
+		return false;
 	switch (cmd) {
 	case 0:
 		slog_debug(server, "S: auth ok");
@@ -285,16 +314,14 @@ bool answer_authreq(PgSocket *server, PktHdr *pkt)
 		break;
 	case 4:
 		slog_debug(server, "S: req crypt psw");
-		if (mbuf_avail(&pkt->data) < 2)
+		if (!mbuf_get_bytes(&pkt->data, 2, &salt))
 			return false;
-		salt = mbuf_get_bytes(&pkt->data, 2);
 		res = login_crypt_psw(server, salt);
 		break;
 	case 5:
 		slog_debug(server, "S: req md5-crypted psw");
-		if (mbuf_avail(&pkt->data) < 4)
+		if (!mbuf_get_bytes(&pkt->data, 4, &salt))
 			return false;
-		salt = mbuf_get_bytes(&pkt->data, 4);
 		res = login_md5_psw(server, salt);
 		break;
 	case 2: /* kerberos */
@@ -314,37 +341,49 @@ bool send_startup_packet(PgSocket *server)
 {
 	PgDatabase *db = server->pool->db;
 	const char *username = server->pool->user->name;
-	PktBuf pkt;
-	uint8_t buf[512];
+	PktBuf *pkt;
 
-	pktbuf_static(&pkt, buf, sizeof(buf));
-	pktbuf_write_StartupMessage(&pkt, username,
-				    db->startup_params,
-				    db->startup_params_len);
-	return pktbuf_send_immidiate(&pkt, server);
+	pkt = pktbuf_temp();
+	pktbuf_write_StartupMessage(pkt, username,
+				    db->startup_params->buf,
+				    db->startup_params->write_pos);
+	return pktbuf_send_immidiate(pkt, server);
 }
 
-int scan_text_result(MBuf *pkt, const char *tupdesc, ...)
+int scan_text_result(struct MBuf *pkt, const char *tupdesc, ...)
 {
-	char *val = NULL;
-	int len;
-	unsigned ncol, i, asked;
+	const char *val = NULL;
+	uint32_t len;
+	uint16_t ncol;
+	unsigned i, asked;
 	va_list ap;
 	int *int_p;
 	uint64_t *long_p;
-	char **str_p;
+	const char **str_p;
 
 	asked = strlen(tupdesc);
-	ncol = mbuf_get_uint16(pkt);
+	if (!mbuf_get_uint16be(pkt, &ncol))
+		return -1;
 
 	va_start(ap, tupdesc);
 	for (i = 0; i < asked; i++) {
 		if (i < ncol) {
-			len = mbuf_get_uint32(pkt);
-			if (len < 0)
+			if (!mbuf_get_uint32be(pkt, &len))
+				return -1;
+			if ((int32_t)len < 0) {
 				val = NULL;
-			else
-				val = (char *)mbuf_get_bytes(pkt, len);
+			} else {
+				if (!mbuf_get_chars(pkt, len, &val))
+					return -1;
+			}
+
+			/* hack to zero-terminate the result */
+			if (val) {
+				char *xval = (char *)val - 1;
+				memmove(xval, val, len);
+				xval[len] = 0;
+				val = xval;
+			}
 		} else
 			/* tuple was shorter than requested */
 			val = NULL;
@@ -352,14 +391,14 @@ int scan_text_result(MBuf *pkt, const char *tupdesc, ...)
 		switch (tupdesc[i]) {
 		case 'i':
 			int_p = va_arg(ap, int *);
-			*int_p = atoi(val);
+			*int_p = val ? atoi(val) : 0;
 			break;
 		case 'q':
 			long_p = va_arg(ap, uint64_t *);
-			*long_p = atoll(val);
+			*long_p = val ? atoll(val) : 0;
 			break;
 		case 's':
-			str_p = va_arg(ap, char **);
+			str_p = va_arg(ap, const char **);
 			*str_p = val;
 			break;
 		default:
