@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -35,6 +35,12 @@
 #define ACT_SKIP 2
 #define ACT_CALL 3
 
+enum WaitType {
+	W_NONE = 0,
+	W_CONNECT,
+	W_RECV,
+	W_SEND,
+};
 
 #define AssertSanity(sbuf) do { \
 	Assert(iobuf_sane((sbuf)->io)); \
@@ -80,16 +86,14 @@ bool sbuf_accept(SBuf *sbuf, int sock, bool is_unix)
 	Assert(iobuf_empty(sbuf->io) && sbuf->sock == 0);
 	AssertSanity(sbuf);
 
-	tune_socket(sock, is_unix);
 	sbuf->sock = sock;
-	sbuf->is_unix = is_unix;
+	if (!tune_socket(sock, is_unix))
+		goto failed;
 
 	if (!cf_reboot) {
 		res = sbuf_wait_for_data(sbuf);
-		if (!res) {
-			sbuf_call_proto(sbuf, SBUF_EV_RECV_FAILED);
-			return false;
-		}
+		if (!res)
+			goto failed;
 		/* socket should already have some data (linux only) */
 		if (cf_tcp_defer_accept && !is_unix) {
 			sbuf_main_loop(sbuf, DO_RECV);
@@ -98,58 +102,39 @@ bool sbuf_accept(SBuf *sbuf, int sock, bool is_unix)
 		}
 	}
 	return true;
+failed:
+	sbuf_call_proto(sbuf, SBUF_EV_RECV_FAILED);
+	return false;
 }
 
 /* need to connect() to get a socket */
-bool sbuf_connect(SBuf *sbuf, const PgAddr *addr, const char *unix_dir, int timeout_sec)
+bool sbuf_connect(SBuf *sbuf, const struct sockaddr *sa, int sa_len, int timeout_sec)
 {
-	int res, sock, domain;
-	struct sockaddr_in sa_in;
-	struct sockaddr_un sa_un;
-	struct sockaddr *sa;
-	socklen_t len;
+	int res, sock;
 	struct timeval timeout;
+	bool is_unix = sa->sa_family == AF_UNIX;
 
 	Assert(iobuf_empty(sbuf->io) && sbuf->sock == 0);
 	AssertSanity(sbuf);
 
-	/* prepare sockaddr */
-	if (addr->is_unix) {
-		sa = (void*)&sa_un;
-		len = sizeof(sa_un);
-		memset(sa, 0, len);
-		sa_un.sun_family = AF_UNIX;
-		snprintf(sa_un.sun_path, sizeof(sa_un.sun_path),
-			 "%s/.s.PGSQL.%d", unix_dir, addr->port);
-		domain = AF_UNIX;
-	} else {
-		sa = (void*)&sa_in;
-		len = sizeof(sa_in);
-		memset(sa, 0, len);
-		sa_in.sin_family = AF_INET;
-		sa_in.sin_addr = addr->ip_addr;
-		sa_in.sin_port = htons(addr->port);
-		domain = AF_INET;
-	}
-
 	/*
 	 * common stuff
 	 */
-	sock = socket(domain, SOCK_STREAM, 0);
+	sock = socket(sa->sa_family, SOCK_STREAM, 0);
 	if (sock < 0)
 		/* probably fd limit */
 		goto failed;
 
-	tune_socket(sock, addr->is_unix);
+	if (!tune_socket(sock, is_unix))
+		goto failed;
 
-	sbuf->is_unix = addr->is_unix;
 	sbuf->sock = sock;
 
 	timeout.tv_sec = timeout_sec;
 	timeout.tv_usec = 0;
 
 	/* launch connection */
-	res = safe_connect(sock, sa, len);
+	res = safe_connect(sock, sa, sa_len);
 	if (res == 0) {
 		/* unix socket gives connection immidiately */
 		sbuf_connect_cb(sock, EV_WRITE, sbuf);
@@ -158,8 +143,10 @@ bool sbuf_connect(SBuf *sbuf, const PgAddr *addr, const char *unix_dir, int time
 		/* tcp socket needs waiting */
 		event_set(&sbuf->ev, sock, EV_WRITE, sbuf_connect_cb, sbuf);
 		res = event_add(&sbuf->ev, &timeout);
-		if (res >= 0)
+		if (res >= 0) {
+			sbuf->wait_type = W_CONNECT;
 			return true;
+		}
 	}
 
 failed:
@@ -176,12 +163,13 @@ failed:
 bool sbuf_pause(SBuf *sbuf)
 {
 	AssertActive(sbuf);
-	Assert(sbuf->wait_send == 0);
+	Assert(sbuf->wait_type == W_RECV);
 
 	if (event_del(&sbuf->ev) < 0) {
 		log_warning("event_del: %s", strerror(errno));
 		return false;
 	}
+	sbuf->wait_type = W_NONE;
 	return true;
 }
 
@@ -233,26 +221,35 @@ bool sbuf_continue_with_callback(SBuf *sbuf, sbuf_libevent_cb user_cb)
 		log_warning("sbuf_continue_with_callback: %s", strerror(errno));
 		return false;
 	}
+	sbuf->wait_type = W_RECV;
 	return true;
 }
 
-/* socket cleanup & close */
+/* socket cleanup & close: keeps .handler and .arg values */
 bool sbuf_close(SBuf *sbuf)
 {
-	/* keep handler & arg values */
-	if (sbuf->sock > 0) {
+	if (sbuf->wait_type) {
+		Assert(sbuf->sock);
+		/* event_del() acts funny occasionally, debug it */
+		errno = 0;
 		if (event_del(&sbuf->ev) < 0) {
-			log_warning("event_del: %s", strerror(errno));
-			return false;
+			if (errno)
+				log_warning("event_del: %s", strerror(errno));
+			else
+				log_warning("event_del: libevent error");
+
+			/* we can retry whole sbuf_close() if needed */
+			/* if (errno == ENOMEM) return false; */
 		}
-		safe_close(sbuf->sock);
 	}
+	if (sbuf->sock > 0)
+		safe_close(sbuf->sock);
 	sbuf->dst = NULL;
 	sbuf->sock = 0;
 	sbuf->pkt_remain = 0;
-	sbuf->pkt_action = sbuf->wait_send = 0;
+	sbuf->pkt_action = sbuf->wait_type = 0;
 	if (sbuf->io) {
-		obj_free(iobuf_cache, sbuf->io);
+		slab_free(iobuf_cache, sbuf->io);
 		sbuf->io = NULL;
 	}
 	return true;
@@ -301,7 +298,7 @@ void sbuf_prepare_fetch(SBuf *sbuf, unsigned amount)
  *************************/
 
 /*
- * Call proto callback with proper MBuf.
+ * Call proto callback with proper struct MBuf.
  *
  * If callback returns true it used one of sbuf_prepare_* on sbuf,
  * and processing can continue.
@@ -312,7 +309,7 @@ void sbuf_prepare_fetch(SBuf *sbuf, unsigned amount)
  */
 static bool sbuf_call_proto(SBuf *sbuf, int event)
 {
-	MBuf mbuf;
+	struct MBuf mbuf;
 	IOBuf *io = sbuf->io;
 	bool res;
 
@@ -346,6 +343,7 @@ static bool sbuf_wait_for_data(SBuf *sbuf)
 		log_warning("sbuf_wait_for_data: event_add: %s", strerror(errno));
 		return false;
 	}
+	sbuf->wait_type = W_RECV;
 	return true;
 }
 
@@ -360,10 +358,11 @@ static void sbuf_send_cb(int sock, short flags, void *arg)
 		return;
 
 	AssertSanity(sbuf);
-	Assert(sbuf->wait_send);
+	Assert(sbuf->wait_type == W_SEND);
+
+	sbuf->wait_type = W_NONE;
 
 	/* prepare normal situation for sbuf_main_loop */
-	sbuf->wait_send = 0;
 	res = sbuf_wait_for_data(sbuf);
 	if (res) {
 		/* here we should certainly skip recv() */
@@ -378,11 +377,13 @@ static bool sbuf_queue_send(SBuf *sbuf)
 {
 	int err;
 	AssertActive(sbuf);
+	Assert(sbuf->wait_type == W_RECV);
 
 	/* if false is returned, the socket will be closed later */
 
 	/* stop waiting for read events */
 	err = event_del(&sbuf->ev);
+	sbuf->wait_type = W_NONE; /* make sure its called only once */
 	if (err < 0) {
 		log_warning("sbuf_queue_send: event_del failed: %s", strerror(errno));
 		return false;
@@ -395,8 +396,8 @@ static bool sbuf_queue_send(SBuf *sbuf)
 		log_warning("sbuf_queue_send: event_add failed: %s", strerror(errno));
 		return false;
 	}
+	sbuf->wait_type = W_SEND;
 
-	sbuf->wait_send = 1;
 	return true;
 }
 
@@ -524,7 +525,7 @@ static void sbuf_try_resync(SBuf *sbuf, bool release)
 		return;
 
 	if (release && iobuf_empty(io)) {
-		obj_free(iobuf_cache, io);
+		slab_free(iobuf_cache, io);
 		sbuf->io = NULL;
 	} else
 		iobuf_try_resync(io, SBUF_SMALL_PKT);
@@ -563,7 +564,7 @@ static void sbuf_recv_cb(int sock, short flags, void *arg)
 static bool allocate_iobuf(SBuf *sbuf)
 {
 	if (sbuf->io == NULL) {
-		sbuf->io = obj_alloc(iobuf_cache);
+		sbuf->io = slab_alloc(iobuf_cache);
 		if (sbuf->io == NULL) {
 			sbuf_call_proto(sbuf, SBUF_EV_RECV_FAILED);
 			return false;
@@ -585,13 +586,14 @@ static bool allocate_iobuf(SBuf *sbuf)
 static void sbuf_main_loop(SBuf *sbuf, bool skip_recv)
 {
 	unsigned free, ok;
+	int loopcnt = 0;
 
 	/* sbuf was closed before in this event loop */
 	if (!sbuf->sock)
 		return;
 
 	/* reading should be disabled when waiting */
-	Assert(sbuf->wait_send == 0);
+	Assert(sbuf->wait_type == W_RECV);
 	AssertSanity(sbuf);
 
 	if (!allocate_iobuf(sbuf))
@@ -605,6 +607,19 @@ try_more:
 	/* make room in buffer */
 	sbuf_try_resync(sbuf, false);
 
+	/* avoid spending too much time on single socket */
+	if (cf_sbuf_loopcnt > 0 && loopcnt >= cf_sbuf_loopcnt) {
+		log_debug("loopcnt full");
+		/*
+		 * sbuf_process_pending() avoids some data if buffer is full,
+		 * but as we exit processing loop here, we need to retry
+		 * after resync to process all data. (result is ignored)
+		 */
+		ok = sbuf_process_pending(sbuf);
+		return;
+	}
+	loopcnt++;
+
 	/*
 	 * here used to be if (free > SBUF_SMALL_PKT) check
 	 * but with skip_recv switch its should not be needed anymore.
@@ -670,6 +685,9 @@ static void sbuf_connect_cb(int sock, short flags, void *arg)
 {
 	SBuf *sbuf = arg;
 
+	Assert(sbuf->wait_type == W_CONNECT || sbuf->wait_type == W_NONE);
+	sbuf->wait_type = W_NONE;
+
 	if (flags & EV_WRITE) {
 		if (!sbuf_after_connect_check(sbuf))
 			goto failed;
