@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -34,12 +34,10 @@ static bool load_parameter(PgSocket *server, PktHdr *pkt, bool startup)
 	if (incomplete_pkt(pkt))
 		return false;
 
-	key = mbuf_get_string(&pkt->data);
-	val = mbuf_get_string(&pkt->data);
-	if (!key || !val) {
-		disconnect_server(server, true, "broken ParameterStatus packet");
-		return false;
-	}
+	if (!mbuf_get_string(&pkt->data, &key))
+		goto failed;
+	if (!mbuf_get_string(&pkt->data, &val))
+		goto failed;
 	slog_debug(server, "S: param: %s = %s", key, val);
 
 	varcache_set(&server->vars, key, val);
@@ -49,16 +47,24 @@ static bool load_parameter(PgSocket *server, PktHdr *pkt, bool startup)
 		varcache_set(&client->vars, key, val);
 	}
 
-	if (startup)
-		add_welcome_parameter(server->pool, key, val);
+	if (startup) {
+		if (!add_welcome_parameter(server->pool, key, val))
+			goto failed_store;
+	}
 
 	return true;
+failed:
+	disconnect_server(server, true, "broken ParameterStatus packet");
+	return false;
+failed_store:
+	disconnect_server(server, true, "failed to store ParameterStatus");
+	return false;
 }
 
 /* we cannot log in at all, notify clients */
 static void kill_pool_logins(PgPool *pool, PktHdr *errpkt)
 {
-	List *item, *tmp;
+	struct List *item, *tmp;
 	PgSocket *client;
 	const char *level, *msg;
 
@@ -71,7 +77,7 @@ static void kill_pool_logins(PgPool *pool, PktHdr *errpkt)
 		if (!client->wait_for_welcome)
 			continue;
 
-		disconnect_client(client, true, msg);
+		disconnect_client(client, true, "%s", msg);
 	}
 }
 
@@ -80,6 +86,7 @@ static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 {
 	SBuf *sbuf = &server->sbuf;
 	bool res = false;
+	const uint8_t *ckey;
 
 	if (incomplete_pkt(pkt)) {
 		disconnect_server(server, true, "partial pkt in login phase");
@@ -158,10 +165,11 @@ static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 
 	/* ignorable packets */
 	case 'K':		/* BackendKeyData */
-		if (mbuf_avail(&pkt->data) >= BACKENDKEY_LEN)
-			memcpy(server->cancel_key,
-			       mbuf_get_bytes(&pkt->data, BACKENDKEY_LEN),
-			       BACKENDKEY_LEN);
+		if (!mbuf_get_bytes(&pkt->data, BACKENDKEY_LEN, &ckey)) {
+			disconnect_server(server, true, "bad cancel key");
+			return false;
+		}
+		memcpy(server->cancel_key, ckey, BACKENDKEY_LEN);
 		res = true;
 		break;
 
@@ -197,16 +205,14 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 	case 'Z':		/* ReadyForQuery */
 
 		/* if partial pkt, wait */
-		if (mbuf_avail(&pkt->data) == 0)
+		if (!mbuf_get_char(&pkt->data, &state))
 			return false;
-		state = mbuf_get_char(&pkt->data);
 
 		/* set ready only if no tx */
 		if (state == 'I')
 			ready = 1;
 		else if (cf_pool_mode == POOL_STMT) {
-			disconnect_server(server, true,
-					  "Long transactions not allowed");
+			disconnect_server(server, true, "Long transactions not allowed");
 			return false;
 		}
 		break;
@@ -302,7 +308,7 @@ static bool handle_connect(PgSocket *server)
 	bool res = false;
 	PgPool *pool = server->pool;
 
-	fill_local_addr(server, sbuf_socket(&server->sbuf), server->remote_addr.is_unix);
+	fill_local_addr(server, sbuf_socket(&server->sbuf), pga_is_unix(&server->remote_addr));
 
 	if (!statlist_empty(&pool->cancel_req_list)) {
 		slog_debug(server, "use it for pending cancel req");
@@ -321,7 +327,7 @@ static bool handle_connect(PgSocket *server)
 }
 
 /* callback from SBuf */
-bool server_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data)
+bool server_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 {
 	bool res = false;
 	PgSocket *server = container_of(sbuf, PgSocket, sbuf);
@@ -343,7 +349,7 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data)
 		disconnect_client(server->link, false, "unexpected eof");
 		break;
 	case SBUF_EV_READ:
-		if (mbuf_avail(data) < NEW_HEADER_LEN) {
+		if (mbuf_avail_for_read(data) < NEW_HEADER_LEN) {
 			slog_noise(server, "S: got partial header, trying to wait a bit");
 			break;
 		}
