@@ -15,7 +15,7 @@ BOUNCER_LOG=test.log
 BOUNCER_INI=test.ini
 BOUNCER_PID=test.pid
 BOUNCER_PORT=`sed -n '/^listen_port/s/listen_port.*=[^0-9]*//p' $BOUNCER_INI`
-BOUNCER_EXE=./pgbouncer
+BOUNCER_EXE="./pgbouncer"
 
 LOGDIR=log
 NC_PORT=6668
@@ -77,7 +77,7 @@ fw_reject_port() {
 fw_reset() {
 	case `uname` in
 	Linux)
-		sudo iptables -F;;
+		sudo iptables -F OUTPUT;;
 	Darwin)
 		sudo ipfw del 100;;
 	*)
@@ -109,7 +109,7 @@ runtest() {
 	echo -n "`date` running $1 ... "
 	eval $1 >$LOGDIR/$1.log 2>&1
 	if [ $? -eq 0 ]; then
-		echo "SUCCESS"
+		echo "ok"
 	else
 		echo "FAILED"
 	fi
@@ -178,15 +178,18 @@ test_server_login_retry() {
 test_server_connect_timeout_establish() {
 	which nc >/dev/null || return 1
 
-	nc -l -p $NC_PORT >/dev/null &
+	echo nc -q 5 -l $NC_PORT
+	nc -l -q 5 $NC_PORT >/dev/null &
+	sleep 2
 	admin "set query_timeout=3"
 	admin "set server_connect_timeout=2"
 	psql -c "select now()" p2
 	# client will always see query_timeout, need to grep for connect timeout
 	grep "closing because: connect timeout" $BOUNCER_LOG 
-        # didnt seem to die otherwise
+	rc=$?
+	# didnt seem to die otherwise
 	killall nc
-	return $?
+	return $rc
 }
 
 # server_connect_timeout - block with iptables
@@ -407,3 +410,4 @@ done
 
 complete
 
+# vim: sts=0 sw=8 noet nosmarttab:
